# This is adapted/translated form ORF3V source code, originaly in 'Go' language

import numpy as np
import math
import random


# Function to calculate Hoeffding bound
def calcHB(delta, windowSize):
    return math.sqrt(math.log(1/delta) / (2 * float(windowSize)))

# Custom queue implementation that pops elements automatically, when pushed into a full queue
class Queue:
    def __init__(self, n):
        self.n = n
        self.queue = []
    
    def add(self, val):
        self.queue.append(val)
        if(len(self.queue) > self.n):
            self.queue.pop(0)
    
    def mean(self):
        if len(self.queue) == 0:
            return 0.0
        m  = 1.0 * sum(self.queue)/len(self.queue)
        return m

# Custom class to store and update feature statistics
class FeatStats:
    def __init__(self, n, x=None, y=None, numClasses = 2):
        self.slidingWindow = Queue(n)
        self.seenInstances = 0
        self.digestPerClass = {i:TDigest() for i in range(numClasses)}  # type: ignore
        self.minValue = None
        self.maxValue = None
        self.classCounts = {i:0.0 for i in range(numClasses)}
        self.update(x, y)

    def update(self, x=None, y=None):
        if( (x is not None) and (y is not None) ):
            y = int(np.squeeze(y))
            self.slidingWindow.add(True)
            self.seenInstances += 1
            self.digestPerClass[y].update(float(np.squeeze(x)))
            
            if (self.minValue is None): self.minValue = x
            else: self.minValue = min(self.minValue, x)
            
            if (self.maxValue is None): self.maxValue = x
            else: self.maxValue = max(self.maxValue, x)

            self.classCounts[y]+=1
        
        else:
            self.slidingWindow.add(False)

# Custiom implementation of decision stump compatible wih TDigest data structure and support update in online fashion
class Stump:
    def __init__(self, minValue, maxValue, digestPerClass, classCounts):
        self.age=0
        self.threshold = minValue + random.random() * (maxValue - minValue)
        self.gini, self.classDistAbove, self.classDistBelow = self.calcApproxGini(digestPerClass, classCounts)
        
        if all(prob == 0.0 for prob in self.classDistAbove.values()):
            self.classDistAbove = {key:1.0/len(self.classDistAbove) for key in self.classDistAbove}
        
        if all(prob == 0.0 for prob in self.classDistBelow.values()):
            self.classDistBelow = {key:1.0/len(self.classDistBelow) for key in self.classDistBelow}

    def calcApproxGini(self, digestPerClass, classCounts):

        totalCount = float(sum(classCounts.values()))
        
        classDistBelow = dict()
        classDistAbove = dict()
        for Class, digest in digestPerClass.items():
            classDistBelow[Class] = digest.cdf(self.threshold)  #P(fi<Xj | C)
            classDistAbove[Class] = 1 - classDistBelow[Class]   #P(fi>Xj | C)
        
        countBelow, countAbove = 0.0, 0.0
        for Class, digest in digestPerClass.items():
            countBelow += classDistBelow[Class] * classCounts[Class] #NB
            countAbove += classDistAbove[Class] * classCounts[Class] #NA
        
        normalizedClassDistBelow, normalizedClassDistAbove = dict(), dict()
        for Class in classCounts:
            normalizedClassDistBelow[Class] = classDistBelow[Class] * classCounts[Class] / (countBelow)\
                                                 if countBelow != 0.0 else 0.0 # P(fi<Xj | C) * (Nc) / NB
            normalizedClassDistAbove[Class] = classDistAbove[Class] * classCounts[Class] / (countAbove)\
                                                 if countAbove != 0.0 else 0.0 # P(fi>Xj | C) * (Nc) / NA
        
        giniBelow, giniAbove = 0.0, 0.0
        for Class in classCounts:
            giniBelow += normalizedClassDistBelow[Class] ** 2
            giniAbove += normalizedClassDistAbove[Class] ** 2
        
        giniGain = 0.0
        if (not np.isnan(giniBelow)) and (not np.isnan(giniAbove)):
            giniGain = (giniAbove * countAbove / totalCount) + (giniBelow * countBelow / totalCount)
        
        return giniGain, normalizedClassDistBelow, normalizedClassDistAbove
    
    def predict(self, x):
        self.age += 1
        if x < self.threshold:
            # return numpy array in class-index order
            return np.array([self.classDistBelow.get(c, 0.0) for c in sorted(self.classDistBelow)])
        else:
            return np.array([self.classDistAbove.get(c, 0.0) for c in sorted(self.classDistAbove)])

# Forest of decision stumps for a particular feature
class FeatureForest:
    def __init__(self, n, feature, featStats, numClasses):
        self.feature = feature
        self.numClasses = numClasses

        minV, maxV = featStats.minValue, featStats.maxValue
        if minV is None or maxV is None:
            minV, maxV = 0.0, 1.0
        
        self.decisionStumps = [
            Stump(
                minValue = minV,
                maxValue = maxV,
                digestPerClass = featStats.digestPerClass,
                classCounts = featStats.classCounts) for _ in range(n)]
        self.weights = [1.0 - random.random()/10000.0 for _ in range(n)] # line #369 from original paper code

    def predict(self, x):
        distAggregate = np.array([0.0]*self.numClasses)
        for stump, weight in zip(self.decisionStumps, self.weights):
            predDist = stump.predict(x)
            distAggregate += predDist * weight
        total_weight = np.sum(distAggregate)
        if total_weight > 0:
            distAggregate = distAggregate/total_weight

        return distAggregate


class ORF3V:
    def __init__(self, Xs, X_masks, Ys, forestSize=200, replacementInterval=50, replacementChance=0.3, windowSize = 200, 
                updateStrategy = "oldest", alpha = 0.1, delta = 0.001, numClasses = 2):
        
        # Check import for optional dependency:
        try:
            from tdigest import TDigest  # type: ignore
            globals()["TDigest"] = TDigest  # make it visible to FeatStats
        except ImportError as e:
            raise ImportError(
            "The 'tdigest' package is required to use the ORF3V model but is not installed.\n\n"
            "To install it along with its dependencies, run:\n\n"
            "    pip install haphazard[orf3v]\n\n"
            "    or\n\n"
            "    pip install tdigest\n\n"
            "If you are on Windows, 'tdigest' requires the Microsoft Visual C++ Build Tools.\n"
            "During installation, select the 'Desktop development with C++' workload and ensure:\n"
            "  • MSVC v143 or later is checked\n"
            "  • Windows 10 or 11 SDK is installed\n"
            "  • CMake tools are included\n\n"
            "Once installed, re-run the above pip command."
        ) from e
        
        # forestSize - Mumber of stumps for every feature
        # replacementInterval - Instances after which stumps might get replaced
        # replacementChance - Probability to not replace stump for "random" update strategy
        # windowsize - Buffer storage size on which to determine feature statistics
        # updateStrategy - Strategy to replace stumps ("oldest" or "random")
        # alpha - Weight update parameter
        # delta - Pruning parameter 
        # num_classes - Number of target classes

        self.forestSize = forestSize
        self.replacementInterval = replacementInterval
        self.replacementChance = replacementChance
        self.windowSize = windowSize
        self.updateStrategy = updateStrategy
        self.alpha = alpha
        self.delta = delta
        self.numClasses = numClasses

        self.counter = 0
        self.firstOccurance = dict()

        self.currentFeatures = set()
        self.featureSet = set()

        self.featStats = dict()
        self.featureForest = dict()
        self.weights = dict()

        self.hb = calcHB(self.delta, self.windowSize)

        self.initialUpdate(Xs, X_masks, Ys)

    def initialUpdate(self, Xs, X_masks, Ys):
        if len(Xs.shape) == 1:
            Xs = Xs.reshape(1, -1)
            X_masks = X_masks.reshape(1, -1)
            Ys = Ys.reshape(1, -1)

        # generate featStats
        for x, x_mask, y in zip(Xs, X_masks, Ys):
            features = set(np.where(x_mask)[0])
            self.featureSet.update(features)
            for feature in features:
                if self.featStats.get(feature, None) is None:
                    self.featStats[feature] = FeatStats(self.windowSize, x[feature], y, self.numClasses)
                else:
                    self.featStats[feature].update(x[feature], y)
        
        # generate featureForests and initalize weights
        for feature in self.featureSet:
            self.firstOccurance[feature] = 0
            self.featureForest[feature] = FeatureForest(self.forestSize, feature,
                                                        self.featStats[feature], self.numClasses)
            self.weights[feature] = 1.0

    def predict(self, X, X_mask):
        self.currentFeatures = set(np.where(X_mask)[0])
        classProbs = np.array([0.0] * self.numClasses)
        for feature in self.currentFeatures:
            if feature in self.featureSet:
                ff = self.featureForest[feature]
                weight = self.weights[feature]

                classProbs += (weight * ff.predict(X[feature]))
        
        if classProbs.sum() > 0:
            classProbs = classProbs / classProbs.sum()
        else:
            classProbs = np.ones(self.numClasses) / float(self.numClasses)

        if self.numClasses == 2:
            return int(np.argmax(classProbs)), float(classProbs[1])
        else:
            return int(np.argmax(classProbs)), classProbs.astype(float).tolist()

    def update(self, X, X_mask, Y):

        '''
        1. Updating the feature statistics,
        2. Pruning of feature forests
        3. Generating feature forests
        4. Updating weights
        5. Replacing decision stumps
        '''

        self.counter += 1
        # Features in Current instance
        self.currentFeatures = set(np.where(X_mask)[0])
        # Feature present in current instance but not in FeatureSet
        newFeatures = self.currentFeatures.difference(self.featureSet)
        # print("New features: ", ne wFeatures)
        # Update featureSet
        self.featureSet.update(newFeatures)

        # Update FeatureStats for existing features
        self.updateFeatureStats(X, Y)

        # Prune FeatureForest for vanished features
        self.pruneForests()

        # Generate new FeatureForest for new features
        self.generateForest(newFeatures, X, Y)

        # Update weights for each feature
        self.updateWeights(X, Y)

        # Replace stupmps from forest according to replacementStrategy
        if self.counter > 10 and self.counter%self.replacementInterval == 0:
            self.replaceStumps()

    def updateFeatureStats(self, X, Y):
        for feature in self.featStats:
            # Feature persent in FeatureSet and current instance
            if feature in self.currentFeatures:
                self.featStats[feature].update(X[feature], Y)
            
            # Feature persent in FeatureSet but not in current instance
            else:
                self.featStats[feature].update()
        return
    
    def pruneForests(self):
        if (self.counter > self.windowSize) and (self.counter%self.replacementInterval == 0):
            pruned_feature = []
            for feature, featStats in self.featStats.items():
                if featStats.seenInstances > self.windowSize:
                    totalMean = (featStats.seenInstances)/(self.counter-self.firstOccurance[feature])
                    windowMean = featStats.slidingWindow.mean()
                    if totalMean-windowMean > self.hb:
                        # self.prune(feature)
                        pruned_feature.append(feature)
            for feature in pruned_feature:
                self.prune(feature)
        return

    def prune(self, feature):
        del self.featureForest[feature], self.weights[feature], self.firstOccurance[feature], self.featStats[feature]
        self.featureSet.remove(feature)
        return  

    def generateForest(self, newFeatures, X, Y):

        for feature in newFeatures:
                self.firstOccurance[feature] = self.counter
                self.weights[feature] = 1.0
                self.featStats[feature] = FeatStats(self.windowSize, X[feature], Y, self.numClasses)
                self.featureForest[feature] = FeatureForest(self.forestSize, feature,
                                                            self.featStats[feature], self.numClasses)
        
        return
    
    def updateWeights(self, X, Y):
        for feature in self.currentFeatures:
            if feature in self.featureForest:
                predDist = self.featureForest[feature].predict(X[feature])
                predClass = np.argmax(predDist)
                self.weights[feature] = (2*self.alpha*(predClass == Y) + self.weights[feature])/(1+self.alpha)
        return

    def replaceStumps(self):
        if self.updateStrategy == "random":
            self.replaceRandomStumps()
        else:
            self.replaceOldestStump()
    
    def replaceRandomStumps(self):
        for feature, ff in self.featureForest.items():
            for i in range(len(ff.decisionStumps)):
                if random.random() < self.replacementChance:
                    ff.decisionStumps[i] = Stump(minValue       =self.featStats[feature].minValue,
                                                  maxValue      =self.featStats[feature].maxValue,
                                                  digestPerClass=self.featStats[feature].digestPerClass,
                                                  classCounts   =self.featStats[feature].classCounts)
        return
    
    def replaceOldestStump(self):
        ageloss = 0.0001*float(self.replacementInterval)
        for feature, ff in self.featureForest.items():
            
            for i, weight in enumerate(ff.weights):
                ff.weights[i] = weight*ageloss
            
            oldestWeight, oldestIdx = 10.0, -1
            for i, weight in enumerate(ff.weights):
                if(weight > oldestWeight):
                    oldestWeight = weight
                    oldestIdx = i
            
            ff.decisionStumps[oldestIdx] = Stump(minValue      =self.featStats[feature].minValue,
                                                 maxValue      =self.featStats[feature].maxValue,
                                                 digestPerClass=self.featStats[feature].digestPerClass,
                                                 classCounts   =self.featStats[feature].classCounts)
            ff.weights[oldestIdx] = 1.0

    def partial_fit(self, X, X_mask, Y):

        Y_pred, Y_logit = self.predict(X, X_mask)
        self.update(X, X_mask, Y)

        return Y_pred, Y_logit


"""
haphazard.normalization.base_normalizer
---------------------------------------
Abstract base class for normalizers.

All normalizer implementations must inherit from `BaseNormalizer` and implement both
`update_params()` and `normalize()` methods. '__init__()` method may be overwritten to
initialize required variables.
"""

from abc import ABC, abstractmethod
from numpy.typing import NDArray

import numpy as np


# -------------------------------------------------------------------------
# BaseNormalizer class
# -------------------------------------------------------------------------
class BaseNormalizer(ABC):
    """
    Abstract Base Class for online (incremental) normalization.

    Each subclass must implement:
        - `update_params()`: update internal running statistics
        - `normalize()`: normalize the observed features

    Notes
    -----
    Works with feature-wise masks, updating only the observed indices.
    """

    def __init__(self, num_features: int, replace_with: float | str = "nan"):
        self.num_features = num_features
        self.replacement = np.nan if replace_with == "nan" else float(replace_with)

    # ------------------------------------------------------------------
    @abstractmethod
    def update_params(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> None:
        """Update internal statistics given the current observed feature indices."""
        raise NotImplementedError

    @abstractmethod
    def normalize(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> NDArray[np.float64]:
        """Normalize observed features using internal statistics."""
        raise NotImplementedError

    # ------------------------------------------------------------------
    def __call__(self, x: NDArray[np.float64], mask: NDArray[np.bool_]) -> NDArray[np.float64]:
        """
        Apply normalization to the observed features in x.

        Parameters
        ----------
        x : NDArray[np.float64]
            Input feature vector of shape (num_features,).
        mask : NDArray[np.bool_]
            Boolean mask where True indicates observed features.

        Returns
        -------
        NDArray[np.float64]
            Normalized vector with missing features replaced.
        """
        indices = np.where(mask)[0]
        if indices.size == 0:
            return np.full_like(x, self.replacement, dtype=np.float64)

        self.update_params(x, indices)
        x_out = self.normalize(x.copy(), indices)
        x_out[~mask] = self.replacement
        return x_out
    
    # ------------------------------------------------------------------
    def reset(self):
        pass


"""
haphazard.models.base_model
---------------------------
Abstract base class for models.

All model implementations must inherit from `BaseModel` and implement `fit()` method 
and call super().__init__() from their constructor.

Notes
-----
- For classification tasks, `fit()` must always return the keys "logits" and "is_logit".
- This ensures metrics computation can safely distinguish between logits and probabilities.
"""

from abc import ABC, abstractmethod
from typing import Any, Literal, Callable
from functools import wraps

import numpy as np

from ..data import BaseDataset


# -------------------------------------------------------------------------
# Type Definitions
# -------------------------------------------------------------------------
TaskType = Literal["classification", "regression"]


# -------------------------------------------------------------------------
# Utility: decorator for runtime validation of fit outputs
# -------------------------------------------------------------------------
def validate_fit_output(func: Callable[..., dict[str, Any]]) -> Callable[..., dict[str, Any]]:
    """
    Decorator to validate that the output of `fit()` matches the expected keys
    and types for the dataset task.
    """
    @wraps(func)
    def wrapper(self, dataset: BaseDataset, *args, **kwargs) -> dict[str, Any]:
        output = func(self, dataset, *args, **kwargs)

        if dataset.task == "classification":
            # Required keys for classification output
            required_keys = {"labels", "preds", "logits", "time_taken", "is_logit"}
            if not all(key in output for key in required_keys):
                missing = required_keys - output.keys()
                raise KeyError(
                    f"{self.__class__.__name__}.fit() missing keys for classification: {missing}"
                )

            # Validate array types
            array_keys = {"labels", "preds", "logits"}
            for key in array_keys:
                if not isinstance(output[key], np.ndarray):
                    raise TypeError(f"'{key}' must be a numpy ndarray, found {type(output[key])!r}")

            # Validate scalar types
            if not isinstance(output["time_taken"], float):
                raise TypeError(f"'time_taken' must be a float, found {type(output['time_taken'])!r}")
            if not isinstance(output["is_logit"], bool):
                raise TypeError(f"'is_logit' must be a boolean, found {type(output['is_logit'])!r}")

        elif dataset.task == "regression":
            # Required keys for regression output
            required_keys = {"targets", "preds", "time_taken"}
            if not all(key in output for key in required_keys):
                missing = required_keys - output.keys()
                raise KeyError(
                    f"{self.__class__.__name__}.fit() missing keys for regression: {missing}"
                )

            # Validate array types
            if not isinstance(output["targets"], np.ndarray):
                raise TypeError(f"'targets' must be a numpy ndarray, found {type(output['targets'])!r}")
            if not isinstance(output["preds"], np.ndarray):
                raise TypeError(f"'preds' must be a numpy ndarray, found {type(output['preds'])!r}")

            # Validate scalar type
            if not isinstance(output["time_taken"], float):
                raise TypeError(f"'time_taken' must be a float, found {type(output['time_taken'])!r}")

        else:
            raise ValueError(f"Unknown dataset.task: {dataset.task}")

        return output

    return wrapper


# -------------------------------------------------------------------------
# BaseModel class
# -------------------------------------------------------------------------
class BaseModel(ABC):
    """
    Abstract base class for all model classes.

    Attributes
    ----------
    name : str
        Name of the model.
    tasks : set[TaskType]
        Supported task types (classification and/or regression).
    deterministic : bool
        Whether the model is deterministic.
    hyperparameters : set[str]
        Set of model hyperparameter names.

    Notes
    -----
    Subclasses must define the following instance attributes **before**
    calling `super().__init__()`:
        - name
        - tasks
        - deterministic
        - hyperparameters
    """

    # Required subclass attributes
    name: str
    tasks: set[TaskType]
    deterministic: bool
    hyperparameters: set[str]

    # ---------------------------------------------------------------------
    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the model runner class.

        Notes
        -----
        - The initializer must accept arbitrary **kwargs.
        - Subclasses must define required attributes before calling this.
        """
        required_attrs = ["name", "tasks", "deterministic", "hyperparameters"]
        for attr in required_attrs:
            if not hasattr(self, attr):
                raise AttributeError(
                    f"{self.__class__.__name__} must define attribute '{attr}' "
                    "before calling super().__init__()."
                )

        # Store the last used hyperparameters for the model
        self.state_dict: dict[str, Any] = {}

    # ---------------------------------------------------------------------
    @abstractmethod
    @validate_fit_output
    def fit(
        self,
        dataset: BaseDataset,
        model_params: dict[str, Any] | None = None,
        seed: int = 42
    ) -> dict[str, Any]:
        """
        Run the model on the given dataset and return results.

        Parameters
        ----------
        dataset : BaseDataset
            Dataset on which the model is trained.
        model_params : dict[str, Any] | None, optional
            Hyperparameters used to initialize the model (default None).
        seed : int, optional
            Random seed for reproducibility (default 42).

        Returns
        -------
        dict[str, Any]
            Classification:
                - labels, NDArray, shape (n_samples,)
                    Ground truth labels.
                - preds, NDArray, shape (n_samples,)
                    Predicted labels.
                - logits, NDArray, shape (n_samples,) or (n_samples, num_classes)
                    Model outputs (logits or probabilities).
                - time_taken, float
                    Time taken to run the model.
                - is_logit, bool
                    True if outputs are raw logits, False if probabilities.

            Regression:
                - targets, NDArray, shape (n_samples,)
                    Ground truth targets.
                - preds, NDArray, shape (n_samples,)
                    Predicted values.
                - time_taken, float
                    Time taken to run the model.
        """
        raise NotImplementedError("Subclasses must implement `fit` method.")

    # ---------------------------------------------------------------------
    def __call__(
        self,
        dataset: BaseDataset,
        model_params: dict[str, Any] | None = None,
        seed: int = 42
    ) -> dict[str, Any]:
        """
        Run the model on the given dataset (wrapper over `fit()`).

        Parameters
        ----------
        dataset : BaseDataset
            Dataset on which the model is trained.
        model_params : dict[str, Any] | None, optional
            Hyperparameters used to initialize the model (default None).
        seed : int, optional
            Random seed for reproducibility (default 42).

        Returns
        -------
        dict[str, Any]
            Model outputs, same format as `fit()`.

        Notes
        -----
        - Uses the last stored `state_dict` if `model_params` is None.
        """
        # Use stored hyperparameters if none provided
        if model_params is None:
            model_params = self.state_dict

        # Ensure required hyperparameters are provided
        if self.hyperparameters and not model_params:
            raise ValueError(
                f"Model '{self.name}' requires hyperparameters {self.hyperparameters}, "
                "but none were provided."
            )

        # Store hyperparameters for future calls
        self.state_dict = model_params

        return self.fit(dataset, self.state_dict, seed)

    # ---------------------------------------------------------------------
    def __repr__(self) -> str:
        """Return a concise string representation of the model."""
        return (
            f"{self.__class__.__name__}("
            f"name='{self.name}', "
            f"tasks={self.tasks}, "
            f"deterministic={self.deterministic}, "
            f"hyperparameters={self.hyperparameters})"
        )



"""
haphazard
---------
Top-level imports for main API access.
"""

# Package version
__version__ = "1.1.7"

# User-facing APIs
from .data import load_dataset
from .models import load_model

# Optional APIs for contributors
from .data import BaseDataset, register_dataset
from .models import BaseModel, register_model

# -------------------------------------------------------------------------
# Public exports
# -------------------------------------------------------------------------
__all__ = [
    "load_dataset",
    "load_model",
    "BaseDataset",
    "register_dataset",
    "BaseModel",
    "register_model",
]



"""
haphazard.normalization.normalizer_zoo.minmax
---------------------------------------------
MinMaxNormalization, online min-max normalization.

This module implements online (incremental) min-max normalization, scaling
each observed feature to a bounded range [0, 1]. It updates feature-wise
minimum and maximum values incrementally and normalizes observed features
accordingly.

The normalized feature is computed as:
    x_norm = (x - min) / (max - min)
"""

from numpy.typing import NDArray
import numpy as np

from ...base_normalizer import BaseNormalizer
from ...normalizer_zoo import register_normalizer


# ----------------------------------------------------------------------
# Min–Max Normalization
# ----------------------------------------------------------------------
@register_normalizer("minmax")
class MinMaxNormalization(BaseNormalizer):
    """
    Online min-max normalization:
        x_norm = (x - min) / (max - min)

    This normalizer tracks running minimum and maximum values for each
    feature and rescales observed features to the range [0, 1]. Designed
    for streaming or incremental data, where feature distributions may
    evolve over time.

    Notes
    -----
    - Handles observed features independently using feature-wise masks.
    - Clips denominator to avoid division by zero.
    - Optionally clips normalized output to [-3, 3] for outlier control.

    Attributes
    ----------
    min_values : NDArray[np.float64]
        Running minimum value per feature.
    max_values : NDArray[np.float64]
        Running maximum value per feature.
    """

    def __init__(self, num_features: int, replace_with: float | str = "nan"):
        super().__init__(num_features, replace_with)
        self.min_values = np.full(num_features, np.inf, dtype=np.float64)
        self.max_values = np.full(num_features, -np.inf, dtype=np.float64)

    def update_params(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> None:
        """Update feature-wise minimum and maximum values."""
        self.min_values[indices] = np.minimum(self.min_values[indices], x[indices])
        self.max_values[indices] = np.maximum(self.max_values[indices], x[indices])

    def normalize(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> NDArray[np.float64]:
        """Normalize observed features to [0, 1] using current min and max."""
        denom = self.max_values[indices] - self.min_values[indices]
        denom = np.clip(denom, 1e-6, None)
        x[indices] = (x[indices] - self.min_values[indices]) / denom
        return np.clip(x, -3.0, 3.0)

    def reset(self) -> None:
        """Reset all running min and max values."""
        self.min_values.fill(np.inf)
        self.max_values.fill(-np.inf)



"""
haphazard.models.model_zoo.olifl
--------------------------------
Wrapper over OLIFL model for binary and multi-class classification.

Implements `RunOLIFL` runner class and a `MultiClassWrapper` for handling
multi-class tasks using a One-vs-Rest strategy.
"""

import copy
import time
from typing import Any

import numpy as np
from numpy.typing import NDArray

from .olifl import OLIFL
from ...base_model import BaseModel, BaseDataset
from ...model_zoo import register_model


class MultiClassWrapper:
    """
    Wraps a base OLIFL model for multi-class classification using One-vs-Rest (OvR).

    Attributes
    ----------
    num_classes : int
        Number of classes.
    models : list[OLIFL]
        List of independent OLIFL models for each class.
    """

    def __init__(self, base_model: OLIFL, num_classes: int):
        """
        Initialize MultiClassWrapper.

        Parameters
        ----------
        base_model : OLIFL
            Base OLIFL model to be wrapped.
        num_classes : int
            Number of classes.
        """
        self.num_classes: int = num_classes
        self.models: list[OLIFL] = [copy.deepcopy(base_model) for _ in range(num_classes)]

    def fit(
        self,
        x: NDArray[np.float64],
        y: NDArray[np.int64],
        x_hap: NDArray[np.float64],
        mask: NDArray[np.bool_],
    ) -> tuple[list[int], list[list[float]]]:
        """
        Train one OLIFL model per class (OvR) and return predictions and logits.

        Parameters
        ----------
        x : NDArray[np.float64], shape (n_samples, n_features)
            Input features.
        y : NDArray[np.int64], shape (n_samples,)
            Target class labels.
        x_hap : NDArray[np.float64], shape (n_samples, n_features)
            Redundant for signature compatibility.
        mask : NDArray[np.bool_], shape (n_samples, n_features)
            Feature availability mask.

        Returns
        -------
        preds : list[int]
            Predicted class labels.
        logits : list[list[float]]
            Per-class logits for each sample.
        """
        logits: list[list[float]] = []

        for cls_idx in range(self.num_classes):
            model = self.models[cls_idx]
            y_binary = np.where(y == cls_idx, 1, 0)
            _, logits_per_class = model.fit(x, y_binary, x_hap, mask)  # type: ignore
            logits.append(logits_per_class)

        # Stack per-class logits into (n_samples, num_classes)
        logits_array = np.column_stack(logits)
        preds = np.argmax(logits_array, axis=1).tolist()

        return preds, logits_array.tolist()


@register_model("olifl")
class RunOLIFL(BaseModel):
    """
    Runner class for OLIFL model.

    Supports binary and multi-class classification tasks. Multi-class
    classification is handled via One-vs-Rest wrapper.

    Attributes
    ----------
    name : str
        Model name.
    tasks : set[str]
        Supported task types.
    deterministic : bool
        Whether model is deterministic.
    hyperparameters : set[str]
        Supported hyperparameters.
    """

    def __init__(self, **kwargs: Any):
        """
        Initialize the OLIFL runner class.

        Parameters
        ----------
        **kwargs
            Optional parameters forwarded to `BaseModel`.
        """
        self.name = "OLIFL"
        self.tasks = {"classification"}
        self.deterministic = False
        self.hyperparameters = {"C", "option"}

        super().__init__(**kwargs)

    def fit(
        self,
        dataset: BaseDataset,
        model_params: dict[str, Any] | None = None,
        seed: int = 42,
    ) -> dict[str, Any]:
        """
        Run the OLIFL model on a dataset.

        Parameters
        ----------
        dataset : BaseDataset
            Dataset object with `.load_data()` and `.load_mask()`.
        model_params : dict
            Hyperparameters for OLIFL model.
        seed : int, default=42
            Random seed for reproducibility.

        Returns
        -------
        dict[str, Any]
            Dictionary containing:
            - labels : Ground truth labels.
            - preds : Predicted labels.
            - logits : Model output logits or probabilities.
            - time_taken : Time taken for full dataset pass.
            - is_logit : Indicates whether scores are logits (True) or probabilities (False).

        Note
        ----
        The dataset object now provides normalized, masked samples via __iter__ and __getitem__,
        enabling online (per-instance) learning without full-batch loading.
        Slicing and iteration behavior is controlled by BaseDataset.
        """
        # --- Validate task ---
        if dataset.task not in self.tasks:
            raise ValueError(
                f"Model {self.__class__.__name__} does not support {dataset.task}. "
                f"Supported task(s): {self.tasks}"
            )
        
        model_params = model_params or {}

        # Set random seed and initialize base_model
        model_params["seed"] = seed
        base_model: OLIFL = OLIFL(**model_params)

        if dataset.task == "regression":
            raise NotImplementedError("Regression task not supported for OLIFL.")

        elif dataset.task == "classification":
            if dataset.num_classes is None:
                raise ValueError(f"'{dataset.name}.num_classes' cannot be None for classification task.")
            
            if dataset.num_classes == 2:
                model: OLIFL | MultiClassWrapper = base_model
            else:
                print(
                    f"[Running {dataset.num_classes} times "
                    "for multi-class classification (OvR strategy).]"
                )
                model = MultiClassWrapper(base_model, num_classes=dataset.num_classes)
            
            # --- Load normalized data ---
            x_mask_y = [xmy for xmy in dataset]
            x, mask, y = zip(*x_mask_y)
            x, mask, y = map(np.asarray, (x, mask, y))

            # --- Train model ---
            start_time = time.perf_counter()
            preds, logits = model.fit(x, y, x, mask)  # type: ignore
            end_time = time.perf_counter()
            time_taken = end_time - start_time

            # --- Format outputs ---
            labels = np.asarray(y, dtype=np.int64)
            preds = np.asarray(preds, dtype=np.int64)
            logits = np.asarray(logits, dtype=np.float64)

            # --- Sanity checks ---
            if dataset.num_classes == 2:
                assert logits.ndim == 1, (
                    f"Expected logits to be 1D for binary classification, got {logits.shape}."
                )
            else:
                assert logits.ndim == 2, (
                    f"Expected logits to be 2D for multi-class classification, got {logits.shape}."
                )

            is_logit = False  # OLIFL returns probabilities

            return {
                "labels": labels,
                "preds": preds,
                "logits": logits,
                "time_taken": time_taken,
                "is_logit": is_logit,
            }
        
        # Fallback for unsupported task
        raise ValueError(f"Unknown task type: '{dataset.task}'")


__all__ = ["RunOLIFL"]



"""
haphazard.models.model_zoo.haptransformer
-----------------------------------------
Wrapper over HapTransformer model for binary and multi-class classification.

Implements `RunHapTransformer` runner class.
"""

import time
from typing import Any

import numpy as np
from numpy.typing import NDArray

from .haptransformer import hapTransformer
from ...base_model import BaseModel, BaseDataset
from ...model_zoo import register_model
from ....utils import seed_everything


@register_model("haptransformer")
class RunHapTransformer(BaseModel):
    """
    Runner class for HapTransformer model.

    Supports binary and multi-class classification tasks.

    Attributes
    ----------
    name : str
        Model name.
    tasks : set[str]
        Supported task types.
    deterministic : bool
        Whether model is deterministic.
    hyperparameters : set[str]
        Supported hyperparameters.
    """

    def __init__(self, **kwargs: Any):
        """
        Initialize the HapTransformer runner class.

        Parameters
        ----------
        **kwargs
            Optional parameters forwarded to `BaseModel`.
        """
        self.name = "HapTransformer"
        self.tasks = {"classification"}
        self.deterministic = False
        self.hyperparameters = {"hidden_size", "n_heads", "lr"}

        super().__init__(**kwargs)

    def fit(
        self,
        dataset: BaseDataset,
        model_params: dict[str, Any] | None = None,
        seed: int = 42,
    ) -> dict[str, Any]:
        """
        Run the HapTransformer model on a dataset.

        Parameters
        ----------
        dataset : BaseDataset
            Dataset object with `.load_data()` and `.load_mask()`.
        model_params : dict
            Hyperparameters for HapTransformer model.
        seed : int, default=42
            Random seed for reproducibility.

        Returns
        -------
        dict[str, Any]
            Dictionary containing:
            - labels : Ground truth labels.
            - preds : Predicted labels.
            - logits : Model output logits or probabilities.
            - time_taken : Time taken for full dataset pass.
            - is_logit : Indicates whether scores are logits (True) or probabilities (False).

        Note
        ----
        The dataset object now provides normalized, masked samples via __iter__ and __getitem__,
        enabling online (per-instance) learning without full-batch loading.
        Slicing and iteration behavior is controlled by BaseDataset.
        """
        # --- Validate task ---
        if dataset.task not in self.tasks:
            raise ValueError(
                f"Model {self.__class__.__name__} does not support {dataset.task}. "
                f"Supported task(s): {self.tasks}"
            )
        
        model_params = model_params or {}
        model_params["n_class"] = dataset.num_classes
        model_params["n_features"] = dataset.n_features
        model_params["device"] = 'cpu'
        model_params["batch_size"] = 1


        if dataset.task == "regression":
            raise NotImplementedError("Regression task not supported for HapTransformer.")

        elif dataset.task == "classification":
            if dataset.num_classes is None:
                raise ValueError(f"'{dataset.name}.num_classes' cannot be None for classification task.")
            
            # Set random seed and initialize base_model
            seed_everything(seed)
            model: hapTransformer = hapTransformer(**model_params)

            # --- Load normalized data ---
            x_mask_y = [xmy for xmy in dataset]
            x, mask, y = zip(*x_mask_y)
            x, mask, y = map(np.asarray, (x, mask, y))
            
            # --- Train model ---
            start_time = time.perf_counter()
            preds, logits = model.fit(x, y, mask)  
            end_time = time.perf_counter()
            time_taken = end_time - start_time

            # --- Format outputs ---
            labels = np.asarray(y, dtype=np.int64)
            preds = np.asarray(preds, dtype=np.int64)
            logits = np.asarray(logits, dtype=np.float64)

            # --- Sanity checks ---
            if dataset.num_classes == 2:
                assert logits.ndim == 1, (
                    f"Expected logits to be 1D for binary classification, got {logits.shape}."
                )
            else:
                assert logits.ndim == 2, (
                    f"Expected logits to be 2D for multi-class classification, got {logits.shape}."
                )

            is_logit = False  # HapTransformer returns probabilities

            return {
                "labels": labels,
                "preds": preds,
                "logits": logits,
                "time_taken": time_taken,
                "is_logit": is_logit,
            }
        
        # Fallback for unsupported task
        raise ValueError(f"Unknown task type: '{dataset.task}'")


__all__ = ["RunHapTransformer"]



"""
haphazard.utils.metrics
-----------------------
Metrics computation utilities for classification and regression tasks.

This module provides helper functions to evaluate model predictions
for both classification (binary and multi-class) and regression tasks.
It supports standard metrics like accuracy, balanced accuracy, AUROC,
AUPRC, and RMSE, and provides a unified dispatcher for easy evaluation.

Implements
----------
- Classification metrics: accuracy, balanced_accuracy, AUROC, AUPRC, number_of_errors
- Regression metrics: RMSE
- Unified dispatcher: get_all_metrics

Notes
-----
- Classification metrics assume `labels` and `preds` are 1D arrays of shape (n_samples,).
- For multi-class AUROC/AUPRC, logits must be of shape (n_samples, num_classes).
- Regression metrics assume continuous target values.
"""

from typing import Literal, Union

import numpy as np
from numpy.typing import NDArray
from sklearn.metrics import (
    roc_auc_score, accuracy_score, average_precision_score,
    balanced_accuracy_score, mean_squared_error
)
from sklearn.preprocessing import label_binarize
from scipy.special import softmax, expit as sigmoid


# -------------------------------------------------------------------------
# Type Definitions
# -------------------------------------------------------------------------
ArrayLike = Union[list, NDArray]
TaskType = Literal["classification", "regression"]

# -------------------------------------------------------------------------
# Classification Metrics
# -------------------------------------------------------------------------
def balanced_accuracy(true: ArrayLike, preds: ArrayLike) -> float:
    """
    Compute the balanced accuracy score.

    Parameters
    ----------
    true : ArrayLike
        Ground truth labels.
    preds : ArrayLike
        Predicted labels.

    Returns
    -------
    float
        Balanced accuracy score.
    """
    return float(balanced_accuracy_score(true, preds))


def accuracy(true: ArrayLike, preds: ArrayLike) -> float:
    """
    Compute standard accuracy score.

    Parameters
    ----------
    true : ArrayLike
        Ground truth labels.
    preds : ArrayLike
        Predicted labels.

    Returns
    -------
    float
        Accuracy score.
    """
    return float(accuracy_score(true, preds))


def auroc(true: ArrayLike, logits: ArrayLike, num_classes: int, is_logit: bool = True) -> float:
    """
    Compute Area Under the Receiver Operating Characteristic (AUROC).

    Parameters
    ----------
    true : ArrayLike
        Ground truth labels.
    logits : ArrayLike
        Model outputs, either raw logits or probabilities.
    num_classes : int
        Number of classes.
    is_logit : bool, default=True
        If True, applies sigmoid/softmax to convert logits to probabilities.

    Returns
    -------
    float
        AUROC score.

    Raises
    ------
    ValueError
        If num_classes < 2.
    AssertionError
        If logits/probs shape does not match expected dimensions.
    """
    true, logits = np.asarray(true), np.asarray(logits)

    if is_logit:
        probs = sigmoid(logits) if num_classes == 2 else softmax(logits, axis=1)
    else:
        probs = logits

    if num_classes == 2:
        assert probs.ndim == 1, f"Expected shape {(true.shape,)}, got {probs.shape}"
        return float(roc_auc_score(true, probs, average="macro"))
    elif num_classes > 2:
        assert probs.ndim == 2, f"Expected shape {(true.shape[0], num_classes)}, got {probs.shape}"
        true_bin = label_binarize(true, classes=np.unique(true))
        return float(roc_auc_score(true_bin, probs, multi_class="ovr"))
    else:
        raise ValueError(f"Invalid num_classes '{num_classes}', expected >= 2.")


def auprc(true: ArrayLike, logits: ArrayLike, num_classes: int, is_logit: bool = True) -> float:
    """
    Compute Area Under the Precision-Recall Curve (AUPRC).

    Parameters
    ----------
    true : ArrayLike
        Ground truth labels.
    logits : ArrayLike
        Model outputs, either raw logits or probabilities.
    num_classes : int
        Number of classes.
    is_logit : bool, default=True
        If True, applies sigmoid/softmax to convert logits to probabilities.

    Returns
    -------
    float
        AUPRC score.

    Raises
    ------
    ValueError
        If num_classes < 2.
    AssertionError
        If logits/probs shape does not match expected dimensions.
    """
    true, logits = np.asarray(true), np.asarray(logits)

    if is_logit:
        probs = sigmoid(logits) if num_classes == 2 else softmax(logits, axis=1)
    else:
        probs = logits

    if num_classes == 2:
        assert probs.ndim == 1, f"Expected shape {(true.shape,)}, got {probs.shape}"
        return float(average_precision_score(true, probs))
    elif num_classes > 2:
        assert probs.ndim == 2, f"Expected shape {(true.shape[0], num_classes)}, got {probs.shape}"
        true_bin = label_binarize(true, classes=np.unique(true))
        return float(average_precision_score(true_bin, probs, average="macro"))
    else:
        raise ValueError(f"Invalid num_classes '{num_classes}', expected >= 2.")


def number_of_errors(true: ArrayLike, preds: ArrayLike) -> int:
    """
    Return the total number of misclassified samples.

    Parameters
    ----------
    true : ArrayLike
        Ground truth labels.
    preds : ArrayLike
        Predicted labels.

    Returns
    -------
    int
        Number of errors.
    """
    true, preds = np.asarray(true), np.asarray(preds)
    return int(np.sum(true.reshape(-1) != preds.reshape(-1)))


# -------------------------------------------------------------------------
# Regression Metrics
# -------------------------------------------------------------------------
def regression_metrics(true: ArrayLike, preds: ArrayLike) -> dict[str, float]:
    """
    Compute regression metrics (currently RMSE).

    Parameters
    ----------
    true : ArrayLike
        Ground truth continuous values.
    preds : ArrayLike
        Predicted continuous values.

    Returns
    -------
    dict[str, float]
        Dictionary with regression metrics:
        {"rmse": float}
    """
    true, preds = np.asarray(true), np.asarray(preds)
    rmse = float(np.sqrt(mean_squared_error(true, preds)))
    return {"rmse": rmse}


# -------------------------------------------------------------------------
# Unified Metrics Dispatcher
# -------------------------------------------------------------------------
def get_all_metrics(
    true: ArrayLike,
    preds: ArrayLike,
    logits: ArrayLike | None = None,
    *,
    task: TaskType = "regression",
    num_classes: int | None = None,
    is_logit: bool = True
) -> dict[str, float]:
    """
    Compute metrics for either regression or classification tasks.

    Parameters
    ----------
    true : ArrayLike
        Ground truth labels or continuous targets.
    preds : ArrayLike
        Predicted labels or continuous values.
    logits : ArrayLike | None, default=None
        Model output logits or probabilities (required for AUROC/AUPRC).
    task : {"classification", "regression"}, default="regression"
        Task type for metrics computation.
    num_classes : int | None, default=None
        Number of classes (used for multi-class metrics if task="classification").
    is_logit : bool, default=True
        Whether `logits` are raw logits (True) or probabilities (False).

    Returns
    -------
    dict[str, float]
        Dictionary of computed metrics.
        - Classification: "acc", "bal_acc", "num_errors", optionally "auroc", "auprc"
        - Regression: "rmse"

    Raises
    ------
    ValueError
        If task type is unsupported.
    """
    true, preds = np.asarray(true), np.asarray(preds)

    if task == "regression":
        return regression_metrics(true, preds)

    elif task == "classification":
        metrics = {
            "bal_acc": balanced_accuracy(true, preds),
            "acc": accuracy(true, preds),
            "num_errors": number_of_errors(true, preds),
        }
        if logits is not None:
            metrics.update({
                "auroc": auroc(true, logits, num_classes or len(np.unique(true)), is_logit=is_logit),
                "auprc": auprc(true, logits, num_classes or len(np.unique(true)), is_logit=is_logit),
            })
        return metrics

    else:
        raise ValueError(f"Unsupported task type '{task}'. Must be 'classification' or 'regression'.")



import numpy as np
from scipy.stats import norm
from statsmodels.distributions.empirical_distribution import ECDF

class OnlineTransformFunction():
    def __init__(self, cont_indices, ord_indices, X=None, window_size=100):
        """
        Require window_size to be positive integers.

        To initialize the window, 
        for continuous columns, sample standatd normal with mean and variance determined by the first batch of observation;
        for ordinal columns, sample uniformly with replacement among all integers between min and max seen from data.

        """
        self.cont_indices = cont_indices
        self.ord_indices = ord_indices
        p = len(cont_indices)
        self.window_size = window_size
        self.window = np.array([[np.nan for x in range(p)] for y in range(self.window_size)]).astype(np.float64)
        self.update_pos = np.zeros(p).astype(int)
        if X is not None:
            self.partial_fit(X)
        
    def partial_fit(self, X_batch):
        """
        Update the running window used to estimate marginals with the data in X
        """
        # Initialization
        if np.isnan(self.window[0, 0] ):
            # Continuous columns: normal initialization

            # mean_cont = np.nanmean(X_batch[:, self.cont_indices])
            # std_cont = np.nanstd(X_batch[:, self.cont_indices])

            '''
            the above lines of code flags the below warning when run with batch_size=1

            'RuntimeWarning: All-NaN slice encountered min_ord = np.nanmin(X_batch[:, j])'
        
            This was dealt by suppressing all warnings in the original code, however we feel that is not
            an appropriate way to handel this and thus the code is changed slightly to handle the warnings.
            '''

            if np.isnan(X_batch[:, self.cont_indices]).all():
                mean_cont = np.nan
            else:
                mean_cont = np.nanmean(X_batch[:, self.cont_indices])
                std_cont = np.nanstd(X_batch[:, self.cont_indices])
            
            if np.isnan(mean_cont):
                self.window[:, self.cont_indices] = np.random.normal(0, 1, size=(self.window_size, np.sum(self.cont_indices)))
            else:
                self.window[:, self.cont_indices] = np.random.normal(mean_cont, std_cont, size=(self.window_size, np.sum(self.cont_indices)))
            # Ordinal columns: uniform initialization
            for j,loc in enumerate(self.ord_indices):
                if loc:
                    # min_ord = np.nanmin(X_batch[:, j])
                    # max_ord = np.nanmax(X_batch[:,j])
                    '''
                    the above lines of code flags the below warning when run with batch_size=1

                    'RuntimeWarning: All-NaN slice encountered min_ord = np.nanmin(X_batch[:, j])'
                
                    This was dealt by suppressing all warnings in the original code, however we feel that is not
                    an appropriate way to handel this and thus the code is changed slightly to handle the warnings.
                    ''' 
                    if np.isnan(X_batch[:, j]).all():
                        min_ord = np.nan
                        max_ord = np.nan
                    else:
                        min_ord = np.nanmin(X_batch[:, j])
                        max_ord = np.nanmax(X_batch[:,j])

                    if np.isnan(min_ord):
                        self.window[:,j].fill(0)
                    else:
                        self.window[:, j] = np.random.randint(min_ord, max_ord+1, size=self.window_size)
        for row in X_batch:
            for col_num in range(len(row)):
                data = row[col_num]
                if not np.isnan(data):
                    self.window[self.update_pos[col_num], col_num] = data
                    self.update_pos[col_num] += 1 
                    if self.update_pos[col_num] >= self.window_size:
                        self.update_pos[col_num] = 0

    def partial_evaluate_cont_latent(self, X_batch):
        """
        Obtain the latent continuous values corresponding to X_batch 
        """
        X_cont = X_batch[:,self.cont_indices]
        window_cont = self.window[:,self.cont_indices]
        Z_cont = np.empty(X_cont.shape)
        Z_cont[:] = np.nan
        for i in range(np.sum(self.cont_indices)):
            # INPUT THE WINDOW FOR EVERY COLUMN
            missing = np.isnan(X_cont[:,i])
            Z_cont[~missing,i] = self.get_cont_latent(X_cont[~missing,i], window_cont[:,i])
        return Z_cont

    def partial_evaluate_ord_latent(self, X_batch):
        """
        Obtain the latent ordinal values corresponding to X_batch
        """
        X_ord = X_batch[:,self.ord_indices]
        window_ord = self.window[:,self.ord_indices]
        Z_ord_lower = np.empty(X_ord.shape)
        Z_ord_lower[:] = np.nan
        Z_ord_upper = np.empty(X_ord.shape)
        Z_ord_upper[:] = np.nan
        for i in range(np.sum(self.ord_indices)):
            missing = np.isnan(X_ord[:,i])
            # INPUT THE WINDOW FOR EVERY COLUMN
            Z_ord_lower[~missing,i], Z_ord_upper[~missing,i] = self.get_ord_latent(X_ord[~missing,i], window_ord[:,i])
        return Z_ord_lower, Z_ord_upper

    def partial_evaluate_cont_observed(self, Z_batch, X_batch=None):
        """
        Transform the latent continous variables in Z_batch into corresponding observations
        """
        Z_cont = Z_batch[:,self.cont_indices]
        if X_batch is None:
            X_batch = np.zeros(Z_batch.shape) * np.nan
        X_cont = X_batch[:,self.cont_indices]
        X_cont_imp = np.copy(X_cont)
        window_cont = self.window[:,self.cont_indices]
        for i in range(np.sum(self.cont_indices)):
            # if X_batch is not provided, missing will be 1:n
            missing = np.isnan(X_cont[:,i])
            if np.sum(missing)>0:
                X_cont_imp[missing,i] = self.get_cont_observed(Z_cont[missing,i], window_cont[:,i])
        return X_cont_imp

    def partial_evaluate_ord_observed(self, Z_batch, X_batch=None):
        """
        Transform the latent ordinal variables in Z_batch into corresponding observations
        """
        Z_ord = Z_batch[:,self.ord_indices]
        if X_batch is None:
            X_batch = np.zeros(Z_batch.shape) * np.nan
        X_ord = X_batch[:, self.ord_indices]
        X_ord_imp = np.copy(X_ord)
        window_ord = self.window[:,self.ord_indices]
        for i in range(np.sum(self.ord_indices)):
            missing = np.isnan(X_ord[:,i])
            if np.sum(missing)>0:
                X_ord_imp[missing,i] = self.get_ord_observed(Z_ord[missing,i], window_ord[:,i])
        return X_ord_imp

    def get_cont_latent(self, x_batch_obs, window):
        """
        Return the latent variables corresponding to the continuous entries of 
        self.X. Estimates the CDF columnwise with the empyrical CDF
        """
        ecdf = ECDF(window)
        l = len(window)
        q = (l / (l + 1.0)) * ecdf(x_batch_obs)
        q[q==0] = l/(l+1)/2
        if any(q==0):
            print("In get_cont_latent, 0 quantile appears")
        return norm.ppf(q)

    def get_cont_observed(self, z_batch_missing, window):
        """
        Applies marginal scaling to convert the latent entries in Z corresponding
        to continuous entries to the corresponding imputed oberserved value
        """
        quantiles = norm.cdf(z_batch_missing)
        if np.isnan(quantiles[0]):
            quantiles[0] = 0.5
        return np.quantile(window, quantiles)

    def get_ord_latent(self, x_batch_obs, window):
        """
        get the cdf at each point in X_batch
        """
        # the lower endpoint of the interval for the cdf
        ecdf = ECDF(window)
        unique = np.unique(window)
        if unique.shape[0] > 1:
            threshold = np.min(np.abs(unique[1:] - unique[:-1]))/2.0
            z_lower_obs = norm.ppf(ecdf(x_batch_obs - threshold))
            z_upper_obs = norm.ppf(ecdf(x_batch_obs + threshold))
        else:
            z_upper_obs = np.inf
            z_lower_obs = -np.inf
        return z_lower_obs, z_upper_obs

    def get_ord_observed(self, z_batch_missing, window, DECIMAL_PRECISION = 3):
        """
        Gets the inverse CDF of Q_batch
        returns: the Q_batch quantiles of the ordinals seen thus far
        """
        n = len(window)
        x = norm.cdf(z_batch_missing)
        # round to avoid numerical errors in ceiling function
        quantile_indices = np.ceil(np.round((n + 1) * x - 1, DECIMAL_PRECISION))
        quantile_indices = np.clip(quantile_indices, a_min=0,a_max=n-1).astype(int)
        sort = np.sort(window)
        return sort[quantile_indices]


"""
haphazard.utils
---------------
Utility module for the Haphazard package.

This module provides general-purpose utilities that assist with:
- Seeding for reproducibility (`seeding.py`)
- File discovery and path utility (`file_utils.py`)
- Metrics computation (`metrics.py`)

Contributors can import helpers directly from this module for convenience:
>>> from haphazard.utils import seed_everything, find_file, get_all_metrics
"""

from .seeding import seed_everything
from .file_utils import find_file
from .metrics import (
    balanced_accuracy,
    accuracy,
    auroc,
    auprc,
    number_of_errors,
    regression_metrics,
    get_all_metrics,
)


# -------------------------------------------------------------------------
# Public exports
# -------------------------------------------------------------------------
__all__ = [
"seed_everything",
"find_file",
"balanced_accuracy",
"accuracy",
"auroc",
"auprc",
"number_of_errors",
"regression_metrics",
"get_all_metrics",
]



"""
haphazard.normalization.normalizer_zoo.mean_normalization
---------------------------------------------------------
MeanNormalization — online mean-centering.

This module implements online mean normalization (centering), which subtracts
the running mean from each observed feature value. It maintains feature-wise
mean estimates incrementally for streaming or online scenarios.

The normalized feature is computed as:
    x_norm = x - mean
"""

from numpy.typing import NDArray
import numpy as np

from ...base_normalizer import BaseNormalizer
from ...normalizer_zoo import register_normalizer


# ----------------------------------------------------------------------
# Mean Normalization
# ----------------------------------------------------------------------
@register_normalizer("mean")
class MeanNormalization(BaseNormalizer):
    """
    Online mean-centering:
        x_norm = x - mean

    This normalizer tracks the running mean of each feature using incremental
    updates. During normalization, the current mean is subtracted from each
    observed feature, effectively centering the data around zero.

    Attributes
    ----------
    count : np.ndarray
        Running count of observations per feature.
    mean : np.ndarray
        Running mean value for each feature.
    """

    def __init__(self, num_features: int, replace_with: float | str = "nan"):
        super().__init__(num_features, replace_with)
        self.count = np.zeros(num_features)
        self.mean = np.zeros(num_features, dtype=np.float64)

    def update_params(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> None:
        """Incrementally update feature-wise running means."""
        self.count[indices] += 1
        self.mean[indices] += (x[indices] - self.mean[indices]) / self.count[indices]

    def normalize(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> NDArray[np.float64]:
        """Subtract the current mean from observed features."""
        x[indices] = x[indices] - self.mean[indices]
        return x

    def reset(self) -> None:
        """Reset running means and counts."""
        self.count.fill(0)
        self.mean.fill(0.0)



import numpy as np
from scipy.stats import norm
from concurrent.futures import ProcessPoolExecutor
from ..transforms.transform_function import TransformFunction
from .embody import _em_step_body_, _em_step_body, _em_step_body_row
import warnings
warnings.filterwarnings("error")

class ExpectationMaximization():
    def __init__(self):
        return

    def impute_missing(self, X, cont_indices=None, ord_indices=None, threshold=0.01, max_iter=50, max_workers=4, max_ord=5, num_ord_updates=1, verbose=False, seed=1):
        """
        Fits a Gaussian Copula and imputes missing values in X.

        Args:
            X (matrix): data matrix with entries to be imputed
            cont_indices (array): logical, true at indices of the continuous entries
            ord_indices (array): logical, true at indices of the ordinal entries
            threshold (float): the threshold for scaled difference between covariance estimates at which to stop early
            max_iter (int): the maximum number of iterations for copula estimation
            max_workers: the maximum number of workers for parallelism
            max_ord: maximum number of levels in any ordinal for detection of ordinal indices
        Returns:
            X_imp (matrix): X with missing values imputed
            sigma_rearragned (matrix): an estimate of the covariance of the copula
        """
        if cont_indices is None and ord_indices is None:
            # guess the indices from the data
            cont_indices = self.get_cont_indices(X, max_ord)
            ord_indices = ~cont_indices
        self.transform_function = TransformFunction(X, cont_indices, ord_indices) # estimate transformation function
        sigma, Z_imp = self._fit_covariance(X, cont_indices, ord_indices, threshold, max_iter, max_workers, num_ord_updates, verbose, seed)
        # rearrange sigma so it corresponds to the column ordering of X #
        # first few dims are always continuous, after always ordinal
        sigma_rearranged = np.empty(sigma.shape)
        sigma_rearranged[np.ix_(ord_indices,ord_indices)] = sigma[:np.sum(ord_indices),:np.sum(ord_indices)]
        sigma_rearranged[np.ix_(cont_indices,cont_indices)] = sigma[np.sum(ord_indices):,np.sum(ord_indices):]
        sigma_rearranged[np.ix_(cont_indices,ord_indices)] = sigma[np.sum(ord_indices):,:np.sum(ord_indices)]
        sigma_rearranged[np.ix_(ord_indices,cont_indices)] =  sigma_rearranged[np.ix_(cont_indices,ord_indices)].T
        # Rearrange Z_imp so that it's columns correspond to the columns of X
        Z_imp_rearranged = np.empty(X.shape)
        Z_imp_rearranged[:,ord_indices] = Z_imp[:,:np.sum(ord_indices)]
        Z_imp_rearranged[:,cont_indices] = Z_imp[:,np.sum(ord_indices):]
        X_imp = np.empty(X.shape)
        X_imp[:,cont_indices] = self.transform_function.impute_cont_observed(Z_imp_rearranged)
        X_imp[:,ord_indices] = self.transform_function.impute_ord_observed(Z_imp_rearranged)
        print("X_imp:")
        print(X_imp)
        np.savetxt('X_imp_2.txt', X_imp)
        return X_imp, sigma_rearranged

    def _fit_covariance(self, X, cont_indices, ord_indices, threshold=0.01, max_iter=100, max_workers=4, num_ord_updates=1, verbose=False, seed=1):
        """
        Fits the covariance matrix of the gaussian copula using the data 
        in X and returns the imputed latent values corresponding to 
        entries of X and the covariance of the copula

        Args:
            X (matrix): data matrix with entries to be imputed
            cont_indices (array): indices of the continuous entries
            ord_indices (array): indices of the ordinal entries
            threshold (float): the threshold for scaled difference between covariance estimates at which to stop early
            max_iter (int): the maximum number of iterations for copula estimation
            max_workers (positive int): the maximum number of workers for parallelism 

        Returns:
            sigma (matrix): an estimate of the covariance of the copula
            Z_imp (matrix): estimates of latent values
        """
        assert cont_indices is not None or ord_indices is not None
        Z_ord_lower, Z_ord_upper = self.transform_function.get_ord_latent()
        Z_ord = self._init_Z_ord(Z_ord_lower, Z_ord_upper, seed)
        Z_cont = self.transform_function.get_cont_latent()

        Z_imp = np.concatenate((Z_ord,Z_cont), axis=1)
        # mean impute the missing continuous values for the sake of covariance estimation
        Z_imp[np.isnan(Z_imp)] = 0.0
        np.savetxt('Z_zeros.txt', Z_imp)
        # initialize the correlation matrix
        sigma = np.corrcoef(Z_imp, rowvar=False)
        # Latent variable matrix with columns sorted as ordinal, continuous
        Z = np.concatenate((Z_ord, Z_cont), axis=1)
        # track previous sigma for the purpose of early stopping
        prev_sigma = self._project_to_correlation(sigma)

        for i in range(max_iter):
            sigma, Z_imp, Z = self._em_step_1(Z, Z_ord_lower, Z_ord_upper, sigma, max_workers, num_ord_updates)
            sigma = self._project_to_correlation(sigma)
            # stop early if the change in the correlation estimation is below the threshold
            sigmaudpate = self._get_scaled_diff(prev_sigma, sigma)
            if sigmaudpate < threshold:
                if verbose: print('Convergence at iteration '+str(i+1))
                break
            if verbose: print("Copula correlation change ratio: ", np.round(sigmaudpate, 4))
            prev_sigma = sigma
        if verbose and i == max_iter-1: 
            print("Convergence not achieved at maximum iterations")
        np.savetxt('Z_imp_2.txt', Z_imp)
        return sigma, Z_imp

    def _em_step(self, Z, r_lower, r_upper, sigma, max_workers=1, num_ord_updates=1):
        """
        Executes one step of the EM algorithm to update the covariance 
        of the copula

        Args:
            Z (matrix): Latent values
            r_lower (matrix): lower bound on latent ordinals
            r_upper (matrix): upper bound on latent ordinals
            sigma (matrix): correlation estimate
            max_workers (positive int): maximum number of workers for parallelism

        Returns:
            sigma (matrix): an estimate of the covariance of the copula
            Z_imp (matrix): estimates of latent values
            Z (matrix): Updated latent values

        """
        n = Z.shape[0]
        p = Z.shape[1]
        res = []
        divide = n/max_workers * np.arange(max_workers+1)
        divide = divide.astype(int)
        args = [(np.copy(Z[divide[i]:divide[i+1],:]), r_lower [divide[i]:divide[i+1],:], r_upper[divide[i]:divide[i+1],:], sigma, num_ord_updates) \
                               for i in range(max_workers)]
        with ProcessPoolExecutor(max_workers=max_workers) as pool:
            res = pool.map(_em_step_body_, args)
            Z_imp = np.empty((n,p))
            C = np.zeros((p,p))
            for i,(C_divide, Z_imp_divide, Z_divide) in enumerate(res):
                C += C_divide/n
                Z_imp[divide[i]:divide[i+1],:] = Z_imp_divide
                Z[divide[i]:divide[i+1],:] = Z_divide
            sigma = np.cov(Z_imp, rowvar=False) + C
            return sigma, Z_imp, Z

    def _em_step_1 (self, Z, r_lower, r_upper, sigma, max_workers=1, num_ord_updates=1):
        """
        Executes one step of the EM algorithm to update the covariance
        of the copula

        Args:
            Z (matrix): Latent values
            r_lower (matrix): lower bound on latent ordinals
            r_upper (matrix): upper bound on latent ordinals
            sigma (matrix): correlation estimate
            max_workers (positive int): maximum number of workers for parallelism

        Returns:
            sigma (matrix): an estimate of the covariance of the copula
            Z_imp (matrix): estimates of latent values
            Z (matrix): Updated latent values

        """
        C, Z_imp, Z=_em_step_body(Z, r_lower, r_upper, sigma, num_ord_updates)
        sigma = np.cov(Z_imp, rowvar=False) + C
        return sigma, Z_imp, Z

    def _project_to_correlation(self, covariance):
        """
        Projects a covariance to a correlation matrix, normalizing it's diagonal entries
        Args:
            covariance (matrix): a covariance matrix
        Returns:
            correlation (matrix): the covariance matrix projected to a correlation matrix
        """
        # print("In em/expecation_matimization, Covariance Matrix: ", covariance)
        D = np.diagonal(covariance)
        # Values of D can't be negative because, sqrt cannot operate on negative value.
        D = abs(D)
        try:
            D_neg_half = 1.0/np.sqrt(D)
        except RuntimeWarning:
            # print("Entered except block")
            corr = np.full_like(covariance, 1)
        else:
            # print("Entered else block")
            covariance *= D_neg_half
            corr = covariance.T * D_neg_half
        # print("In em/expecation_matimization, Correlation Matrix: ", corr)
        return corr

    def _init_Z_ord(self, Z_ord_lower, Z_ord_upper, seed):
        """
        Initializes the observed latent ordinal values by sampling from a standard
        Gaussian truncated to the inveral of Z_ord_lower, Z_ord_upper

        Args:
            Z_ord_lower (matrix): lower range for ordinals
            Z_ord_upper (matrix): upper range for ordinals

        Returns:
            Z_ord (range): Samples drawn from gaussian truncated between Z_ord_lower and Z_ord_upper
        """
        Z_ord = np.empty(Z_ord_lower.shape)
        Z_ord[:] = np.nan

        n, k = Z_ord.shape
        obs_indices = ~np.isnan(Z_ord_lower)

        u_lower = np.copy(Z_ord_lower)
        u_lower[obs_indices] = norm.cdf(Z_ord_lower[obs_indices])
        u_upper = np.copy(Z_ord_upper)
        u_upper[obs_indices] = norm.cdf(Z_ord_upper[obs_indices])

        np.random.seed(seed)
        for i in range(n):
            for j in range(k):
                if not np.isnan(Z_ord_upper[i,j]) and u_upper[i,j] > 0 and u_lower[i,j]<1:
                    u_sample = np.random.uniform(u_lower[i,j],u_upper[i,j])
                    Z_ord[i,j] = norm.ppf(u_sample)
        return Z_ord

    def _get_scaled_diff(self, prev_sigma, sigma):
        """
        Get's the scaled difference between two correlation matrices

        Args:
            prev_sigma (matrix): previous estimate of a matrix
            sigma (matrix): current estimate of a matrix

        Returns: 
            diff (float): scaled distance between the inputs
        """
        return np.linalg.norm(sigma - prev_sigma) / np.linalg.norm(sigma)

    def get_cont_indices(self, X, max_ord):
        """
        get's the indices of continuos columns by returning
        those indicies which have at least max_ord distinct values

        Args:
            X (matrix): input matrix
            max_ord (int): maximum number of distinct values an ordinal can take on in a column

        Returns:
            indices (array): indices of the columns which have at most max_ord distinct entries
        """
        indices = np.zeros(X.shape[1]).astype(bool)
        for i, col in enumerate(X.T):
            col_nonan = col[~np.isnan(col)]
            col_unique = np.unique(col_nonan)
            if len(col_unique) > max_ord:
                indices[i] = True
        return indices


import numpy as np
from statsmodels.distributions.empirical_distribution import ECDF
from scipy.stats import norm

class TransformFunction():
    def __init__(self, X, cont_indices, ord_indices):
        self.X = X
        self.ord_indices = ord_indices
        self.cont_indices = cont_indices

    def get_cont_latent(self):
        """
        Return the latent variables corresponding to the continuous entries of 
        self.X. Estimates the CDF columnwise with the empyrical CDF
        """
        X_cont = self.X[:,self.cont_indices]
        Z_cont = np.empty(X_cont.shape)
        n = self.X.shape[0]
        for i, x_col in enumerate(X_cont.T):
            missing = np.isnan(x_col)
            x_col_noNan = x_col[~missing]
            ecdf = ECDF(x_col_noNan)
            Z_cont[:,i] = norm(0,0.25).ppf((n / (n + 1.0)) * ecdf(x_col))
            # re-add the nan values
            Z_cont[missing,i] = np.nan
        return Z_cont

    def get_ord_latent(self):
        """
        Return the lower and upper ranges of the latent variables corresponding 
        to the ordinal entries of X. Estimates the CDF columnwise with the empyrical CDF
        """
        X_ord = self.X[:,self.ord_indices]
        Z_ord_lower = np.empty(X_ord.shape)
        Z_ord_upper = np.empty(X_ord.shape)
        for i, x_col in enumerate(X_ord.T):
            missing = np.isnan(x_col)
            x_col_noNan = x_col[~missing]
            ecdf = ECDF(x_col_noNan)
            unique = np.unique(x_col_noNan)
            # half the min differenence between two ordinals
            threshold = np.min(np.abs(unique[1:] - unique[:-1]))/2.0
            Z_ord_lower[:,i] = norm(0,0.25).ppf(ecdf(x_col - threshold))
            Z_ord_upper[:,i] = norm(0,0.25).ppf(ecdf(x_col + threshold))
            # re-add the nan values
            Z_ord_lower[missing,i] = np.nan
            Z_ord_upper[missing,i] = np.nan
        return Z_ord_lower, Z_ord_upper

    def impute_cont_observed(self, Z):
        """
        Applies marginal scaling to convert the latent entries in Z corresponding
        to continuous entries to the corresponding imputed oberserved value
        """
        X_cont = self.X[:, self.cont_indices]
        Z_cont = Z[:, self.cont_indices]
        X_imp = np.copy(X_cont)
        for i, x_col in enumerate(X_cont.T):
            missing = np.isnan(x_col)
            # Only impute missing entries
            X_imp[missing,i] = np.quantile(x_col[~missing], norm.cdf(Z_cont[missing,i]))
        return X_imp

    def impute_ord_observed(self, Z):
        """
        Applies marginal scaling to convert the latent entries in Z corresponding
        to ordinal entries to the corresponding imputed oberserved value
        """
        X_ord = self.X[:, self.ord_indices]
        Z_ord = Z[:, self.ord_indices]
        X_imp = np.copy(X_ord)
        for i, x_col in enumerate(X_ord.T):
            missing = np.isnan(x_col)
            # only impute missing entries
            X_imp[missing,i] = self.inverse_ecdf(x_col[~missing], norm.cdf(Z_ord[missing,i]))
        return X_imp

    def inverse_ecdf(self, data, x, DECIMAL_PRECISION = 3):
        """
        computes the inverse ecdf (quantile) for x with ecdf given by data
        """
        n = len(data)
        # round to avoid numerical errors in ceiling function
        quantile_indices = np.ceil(np.round_((n + 1) * x - 1, DECIMAL_PRECISION))
        quantile_indices = np.clip(quantile_indices, a_min=0,a_max=n-1).astype(int)
        sort = np.sort(data)
        return sort[quantile_indices]


import numpy as np
from scipy.stats import norm, truncnorm
import warnings
warnings.filterwarnings("error")

def _em_step_body_(args):
    """
    Does a step of the EM algorithm, needed to dereference args to support parallelism
    """
    return _em_step_body(*args)

def _em_step_body(Z, r_lower, r_upper, sigma, num_ord_updates=1):
    """
    Iterate the rows over provided matrix
    """
    num, p = Z.shape
    Z_imp = np.copy(Z)
    C = np.zeros((p,p))
    for i in range(num):
        # try:
        #     c, z_imp, z = _em_step_body_row(Z[i,:], r_lower[i,:], r_upper[i,:], sigma)
        # except:
        #     np.savetxt("Z.txt", Z)
        #     print(Z)
        c, z_imp, z = _em_step_body_row(Z[i,:], r_lower[i,:], r_upper[i,:], sigma)
        Z_imp[i,:] = z_imp
        Z[i,:] = z
        C += c
    return C, Z_imp, Z

def _em_step_body_row(Z_row, r_lower_row, r_upper_row, sigma, num_ord_updates=1):
    """
    The body of the em algorithm for each row
    Returns a new latent row, latent imputed row and C matrix, which, when added
    to the empirical covariance gives the expected covariance
    Args:
        Z_row (array): (potentially missing) latent entries for one data point
        r_lower_row (array): (potentially missing) lower range of ordinal entries for one data point
        r_upper_row (array): (potentially missing) upper range of ordinal entries for one data point
        sigma (matrix): estimate of covariance
        num_ord (int): the number of ordinal columns

    Returns:
        C (matrix): results in the updated covariance when added to the empircal covariance
        Z_imp_row (array): Z_row with latent ordinals updated and missing entries imputed 
        Z_row (array): inpute Z_row with latent ordinals updated
    """
    Z_imp_row = np.copy(Z_row)
    p = Z_imp_row.shape[0]
    num_ord = r_upper_row.shape[0]
    C = np.zeros((p,p))

    obs_indices = np.where(~np.isnan(Z_row))[0]

    missing_indices = np.setdiff1d(np.arange(p), obs_indices)
    ord_in_obs = np.where(obs_indices < num_ord)[0]
    ord_obs_indices = obs_indices[ord_in_obs]
    # obtain correlation sub-matrices
    # obtain submatrices by indexing a "cartesian-product" of index arrays
    sigma_obs_obs = sigma[np.ix_(obs_indices,obs_indices)]
    sigma_obs_missing = sigma[np.ix_(obs_indices, missing_indices)]
    sigma_missing_missing = sigma[np.ix_(missing_indices, missing_indices)]


    if len(missing_indices) > 0:
        tot_matrix = np.concatenate((np.identity(len(sigma_obs_obs)), sigma_obs_missing), axis=1)

        # intermed_matrix = np.linalg.solve(sigma_obs_obs, tot_matrix)
        '''
        The above line of code sometime ran into the error stated below.
        'LinAlgError: Singular matrix'
        
        So, to handle the cases when a matrix might become singlar, we take the approximate solution
        instead of the exact solution.        
        '''
        try:
            intermed_matrix = np.linalg.solve(sigma_obs_obs, tot_matrix)
        except np.linalg.LinAlgError:
            intermed_matrix = np.linalg.lstsq(sigma_obs_obs, tot_matrix, rcond=None)[0]
        
        sigma_obs_obs_inv = intermed_matrix[:, :len(sigma_obs_obs)]
        J_obs_missing = intermed_matrix[:, len(sigma_obs_obs):]
    else:
        # LinAlgError: Singular matrix
        # below code is changes slightly to handle the above warning
        
        # sigma_obs_obs_inv = np.linalg.solve(sigma_obs_obs, np.identity(len(sigma_obs_obs)))
        '''
        The above line of code sometime ran into the error stated below.
        'LinAlgError: Singular matrix'
        
        So, to handle the cases when a matrix might become singlar, we take the approximate solution
        instead of the exact solution.        
        '''
        try:
            sigma_obs_obs_inv = np.linalg.solve(sigma_obs_obs, np.identity(len(sigma_obs_obs)))
        except np.linalg.LinAlgError:
            sigma_obs_obs_inv = np.linalg.lstsq(sigma_obs_obs, np.identity(len(sigma_obs_obs)), rcond=None)[0]

    # initialize vector of variances for observed ordinal dimensions
    var_ordinal = np.zeros(p)

    # OBSERVED ORDINAL ELEMENTS
    # when there is an observed ordinal to be imputed and another observed dimension, impute this ordinal
    if len(obs_indices) >= 2 and len(ord_obs_indices) >= 1:
        for update_iter in range(num_ord_updates):
            # used to efficiently compute conditional mean
            sigma_obs_obs_inv_Z_row = np.dot(sigma_obs_obs_inv, Z_row[obs_indices])
            for ind in range(len(ord_obs_indices)):
                j = obs_indices[ind]
                not_j_in_obs = np.setdiff1d(np.arange(len(obs_indices)),ind) 
                v = sigma_obs_obs_inv[:,ind]
                # new_var_ij = np.asscalar(1.0/v[ind])
                '''
                The above line of code is depriciated. Updated code below
                '''
                new_var_ij = 1.0/v[ind]

                new_mean_ij = Z_row[j] - new_var_ij*sigma_obs_obs_inv_Z_row[ind]

                # To make sure new_var_ij is positive, we take absolute.
                new_var_ij = abs(new_var_ij)

                # print("In embody, new_var_ij", new_var_ij)

                try:
                    mean, var = truncnorm.stats(
                        a=(r_lower_row[j] - new_mean_ij) / np.sqrt(new_var_ij),
                        b=(r_upper_row[j] - new_mean_ij) / np.sqrt(new_var_ij),
                        loc=new_mean_ij,
                        scale=np.sqrt(new_var_ij),
                        moments='mv')
                    if np.isfinite(var):
                        var_ordinal[j] = var
                        if update_iter == num_ord_updates - 1:
                            C[j,j] = C[j,j] + var 
                    if np.isfinite(mean):
                        Z_row[j] = mean
                except OverflowError:
                    print("Truncnorm stats gives overflow error because the range is out of tail. We handle this by assuming that mean and var is infinite.")
                except RuntimeWarning:
                    print("Truncnorm stats gives runtime warning becasue of invalid value encountered in power. We handle this by assuming that mean and var is infinite.")
                

    # MISSING ELEMENTS
    Z_obs = Z_row[obs_indices]
    Z_imp_row[obs_indices] = Z_obs
    if len(missing_indices) > 0:
        Z_imp_row[missing_indices] = np.matmul(J_obs_missing.T,Z_obs) 
        # variance expectation and imputation
        if len(ord_obs_indices) >= 1 and len(obs_indices) >= 2 and np.sum(var_ordinal) > 0: 
            cov_missing_obs_ord = J_obs_missing[ord_in_obs].T * var_ordinal[ord_obs_indices]
            C[np.ix_(missing_indices, ord_obs_indices)] += cov_missing_obs_ord
            C[np.ix_(ord_obs_indices, missing_indices)] += cov_missing_obs_ord.T
            C[np.ix_(missing_indices, missing_indices)] += sigma_missing_missing - np.matmul(J_obs_missing.T, sigma_obs_missing) \
                                                           + np.matmul(cov_missing_obs_ord, J_obs_missing[ord_in_obs])
        else:
            C[np.ix_(missing_indices, missing_indices)] += sigma_missing_missing - np.matmul(J_obs_missing.T, sigma_obs_missing)
    return C, Z_imp_row, Z_row


"""
haphazard.models
----------------
Models Subpackage

Provides the `BaseModel` abstract class that all registered models
must inherit from. Also contains the `model_zoo` subpackage which
maintains all registered models and their loaders.
"""

from .base_model import BaseModel, TaskType
from .model_zoo import load_model, _MODEL_REGISTRY, register_model


# -------------------------------------------------------------------------
# Public exports
# -------------------------------------------------------------------------

__all__ = [
    "BaseModel",
    "TaskType",
    "load_model",
    "_MODEL_REGISTRY",
    "register_model"
]



# https://github.com/Rohit102497/HaphazardInputsAsImages/blob/main/Code/Utils/utils.py#L80

import io

import numpy as np 
import pandas as pd 
from PIL import Image 
import matplotlib.pyplot as plt 
import torch 
from torchvision import transforms
import torchvision.transforms.functional as TF
import torch.nn.functional as F

trans=transforms.ToTensor()

#------------Min-Max Normalization----------------------#
# performs min max normalization on supplied data instance with dropped features
# requires running min and max for all features, returnss their updated values
# for the first entry in dataset min max will be intialized to an array consisting of nan values
def minmaxnorm(row, min_arr, max_arr, epsilon=1e-15):
    # Create boolean mask for non-nan values
    valid_mask = ~np.isnan(row)
    
    # Handle initial case for min_arr and max_arr
    min_mask = np.isnan(min_arr)
    max_mask = np.isnan(max_arr)
    min_arr = np.where(min_mask & valid_mask, row, min_arr)
    max_arr = np.where(max_mask & valid_mask, row, max_arr)
    
    # Update min and max arrays
    np.minimum(min_arr, row, out=min_arr, where=valid_mask)
    np.maximum(max_arr, row, out=max_arr, where=valid_mask)
    
    # Perform min-max normalization
    norm_row = np.full_like(row, np.nan)
    denominator = max_arr - min_arr + epsilon
    np.divide(row - min_arr, denominator, out=norm_row, where=valid_mask & (denominator != 0))
    
    return norm_row, min_arr, max_arr

def zscore(row, run_sum, sum_sq, count, epsilon=1e-30):
    mask=~np.isnan(row) # select visible, that is, non-nan values
    
    count+=(mask) # the list count maintains the count of instances encounterd for all features. Increment count for each visible feature
    
    run_sum= np.where(mask,run_sum+row,run_sum) # update running sum for all non-nan features
    sum_sq= np.where(mask, sum_sq+row**2, sum_sq) # update sum of squares for all non-nan features
    
    mean=np.where(mask,run_sum/(count+epsilon),np.nan) # calculate mean for all non-nan features
    std= np.where(mask, np.sqrt(sum_sq/(count+epsilon) - mean**2), np.nan) + epsilon # calculate standard deviation for all non-nan features. Add epsilon to avoid division by zero
    
    norm_row= np.where(mask, (row-mean)/(std+epsilon), np.nan) # calculate z-score normalized values for all non-nan features
    
    return norm_row, run_sum, sum_sq, count


#------------------------- Nan Marking -----------------------#

# NAN VALUES REPRESENTED AS CROSSES
# function to produce image tensors corresponding to given normalized data instance. USES MATPLOTLIB AND HENCE IS VERY SLOW. 
# takes normalized and dropped values, reverse mask, color scheme and dots per inch resolution
def bar_nan_mark_z_score_plot(values,rev_val,colors,feat,vert=True,dpi=56):
    s=224/dpi                              #s is the size of image in inches. Models require height and width of image to be 224
    fig, _=plt.subplots(figsize=(s,s))     #fix fig as s*s image. When saving with corresponding dpi we get 224*224 image
    fig=plt.bar(feat,values,color=colors)   #represnt the min-max normed features as bars
    fig=plt.scatter(feat,rev_val,s=180,color='red',marker='x')  #plot crosses at dropped instances
    fig=plt.ylim((-3,3))                     #set y axis limits so that all graphs are scaled uniformly
    fig=io.BytesIO()                        #creates byte buffer for converting plt bar object to jpg
    
    figsav=plt.savefig(fig,dpi=dpi, format='jpg') #saving to jpg
    
    image=trans(Image.open(fig))         #opening jpg using PIL and transforming to PyTorch tensor  
    if not vert:
        image=TF.rotate(image,angle=270)
    plt.close()
    return image

#-----------------------------No-Nan Markings---------------------#
def bar_min_max_plot(values, colors, spacing=0.3):# for min max normalized data
    # Convert inputs to PyTorch tensors if they're not already
    values = torch.tensor(values)
    colors = torch.tensor(colors)

    # Filter out NaN values and their corresponding colors
    valid_mask = ~torch.isnan(values)
    valid_values = values[valid_mask]
    valid_colors = colors[valid_mask]

    # Create a 3x224x224 RGB image with white background
    img = torch.ones((3, 224, 224), dtype=torch.float32)

    num_bars = len(valid_values)
    bar_width = int(224 / (num_bars + (num_bars + 1) * spacing))
    gap_width = int(bar_width * spacing)

    # Ensure we have at least 1 pixel for each bar and gap
    if bar_width == 0 or gap_width == 0:
        bar_width = max(1, int(224 / (2 * num_bars - 1)))
        gap_width = max(1, 224 - num_bars * bar_width) // (num_bars + 1)

    # Calculate positions for bars
    positions = torch.arange(num_bars) * (bar_width + gap_width) + gap_width

    # Calculate heights for bars
    heights = (valid_values * 224).int()

    # Create mask for all bars at once
    y_indices = torch.arange(224).unsqueeze(1).unsqueeze(2)
    x_indices = torch.arange(224).unsqueeze(0).unsqueeze(2)
    
    positions = positions.unsqueeze(0).unsqueeze(0)
    heights = heights.unsqueeze(0).unsqueeze(0)

    bar_masks = (x_indices >= positions) & (x_indices < (positions + bar_width)) & (y_indices >= (224 - heights))

    # Apply colors to bars
    for i, color in enumerate(valid_colors):
        img[:, bar_masks[:, :, i]] = color.float().unsqueeze(1)

    return img


def pie_min_max_plot(sizes, colors, max_sectors=1000):# plots pie charts without using matplotlib.
    sizes=torch.Tensor(sizes)
    colors=torch.Tensor(colors)
    # Filter out NaN values and their corresponding colors
    valid_mask = ~torch.isnan(sizes)
    valid_sizes = sizes[valid_mask]
    valid_colors = colors[valid_mask].to(torch.float32)
    
    # Return blank image in the edge case that all features have 0 values
    if valid_sizes.sum() == 0:
        return torch.ones((3, 224, 224), dtype=torch.float32)
    
    num_sectors = len(valid_sizes)
    
    if num_sectors <= max_sectors:
        # Original method
        img = torch.ones((3, 224, 224), dtype=torch.float32)
        center = torch.tensor([112, 112])
        radius = 100
        
        total = valid_sizes.sum()
        angles = (valid_sizes / total) * (2 * torch.pi)
        cumulative_angles = torch.cumsum(angles, dim=0)
        
        y, x = torch.meshgrid(torch.arange(224), torch.arange(224), indexing='ij')
        x = x - center[0]
        y = y - center[1]
        
        r = torch.sqrt(x**2 + y**2)
        theta = torch.atan2(y, x)
        theta = torch.where(theta < 0, theta + 2*torch.pi, theta)
        
        masks = [(r <= radius) & (theta >= prev_angle) & (theta < curr_angle) 
                 for prev_angle, curr_angle in zip(torch.cat([torch.tensor([0.]), cumulative_angles[:-1]]), cumulative_angles)]
        
        for mask, color in zip(masks, valid_colors):
            img[:, mask] = color.unsqueeze(1)
    
    else:
        # Resizing method for large number of sectors
        size = 1000
        img = torch.ones((3, size, size), dtype=torch.float32)
        center = torch.tensor([size/2, size/2])
        radius = size/2 - 1
        
        # Sort and limit the number of sectors if it exceeds max_sectors
        sorted_indices = torch.argsort(valid_sizes, descending=True)
        valid_sizes = valid_sizes[sorted_indices[:max_sectors-1]]
        valid_colors = valid_colors[sorted_indices[:max_sectors-1]]
        others_size = sizes[sorted_indices[max_sectors-1:]].sum()
        valid_sizes = torch.cat([valid_sizes, others_size.unsqueeze(0)])
        valid_colors = torch.cat([valid_colors, torch.tensor([[0.5, 0.5, 0.5]], dtype=torch.float32)])
        
        total = valid_sizes.sum()
        angles = (valid_sizes / total) * (2 * torch.pi)
        cumulative_angles = torch.cumsum(angles, dim=0)
        
        y, x = torch.meshgrid(torch.arange(size), torch.arange(size), indexing='ij')
        x = x - center[0]
        y = y - center[1]
        
        r = torch.sqrt(x**2 + y**2)
        theta = torch.atan2(y, x)
        theta = torch.where(theta < 0, theta + 2*torch.pi, theta)
        
        for i, (start_angle, end_angle, color) in enumerate(zip(torch.cat([torch.tensor([0.]), cumulative_angles[:-1]]), cumulative_angles, valid_colors)):
            mask = (r <= radius) & (theta >= start_angle) & (theta < end_angle)
            img[:, mask] = color.unsqueeze(1)
        
        img = img.unsqueeze(0)
        img = F.interpolate(img, size=(224, 224), mode='bilinear', align_corners=False)
        img = img.squeeze(0)

    return img

# plots values clipped in [-3,3] as bars. Positive values are plotted above x-axis and negative values below x-axis
def bar_z_score_plot(values, colors, spacing=0.3):
    # Convert inputs to PyTorch tensors if they're not already
    values = torch.tensor(values)
    colors = torch.tensor(colors, dtype=torch.float32)

    # Filter out NaN values and their corresponding colors
    valid_mask = ~torch.isnan(values)
    valid_values = values[valid_mask]
    valid_colors = colors[valid_mask]

    num_bars = len(valid_values)
    
    # Use original method if possible
    bar_width = int(224 / (num_bars + (num_bars + 1) * spacing))
    gap_width = int(bar_width * spacing)
    
    if bar_width >= 1 and gap_width >= 1:
        # Original method
        img = torch.ones((3, 224, 224), dtype=torch.float32)
        
        positions = torch.arange(num_bars) * (bar_width + gap_width) + gap_width
        
        # Calculate heights for positive and negative values separately
        heights = (torch.abs(valid_values) / 3 * 112).int().clamp(0, 112)

        y_indices = torch.arange(224).unsqueeze(1).unsqueeze(2)
        x_indices = torch.arange(224).unsqueeze(0).unsqueeze(2)
        
        positions = positions.unsqueeze(0).unsqueeze(0)
        heights = heights.unsqueeze(0).unsqueeze(0)
        
        # Create masks for positive and negative bars
        positive_mask = valid_values >= 0
        negative_mask = valid_values < 0
        
        positive_bar_masks = (x_indices >= positions) & (x_indices < (positions + bar_width)) & (y_indices >= (112 - heights)) & (y_indices < 112) & positive_mask
        negative_bar_masks = (x_indices >= positions) & (x_indices < (positions + bar_width)) & (y_indices >= 112) & (y_indices < (112 + heights)) & negative_mask
        
        for i, color in enumerate(valid_colors):
            img[:, positive_bar_masks[:, :, i]] = color.unsqueeze(1)
            img[:, negative_bar_masks[:, :, i]] = color.unsqueeze(1)
        
        # Add a horizontal line at y=0
        img[:, 111:113, :] = torch.tensor([0, 0, 0]).unsqueeze(1).unsqueeze(2).expand(3, 2, 224)  # Black line
    
    else:
        # Resizing method for large number of bars
        max_width = max(2000, num_bars * 2)  # Ensure at least 2 pixels per bar
        bar_width = max(1, int(max_width / (num_bars + (num_bars + 1) * spacing)))
        gap_width = max(1, int(bar_width * spacing))
        total_width = num_bars * bar_width + (num_bars + 1) * gap_width
        
        img = torch.ones((3, 224, total_width), dtype=torch.float32)
        positions = torch.arange(num_bars) * (bar_width + gap_width) + gap_width
        
        # Calculate heights for positive and negative values separately
        heights = (torch.abs(valid_values) / 3 * 112).int().clamp(0, 112)

        for i, (pos, height, color, value) in enumerate(zip(positions, heights, valid_colors, valid_values)):
            if value >= 0:
                img[:, 112-height:112, pos:pos+bar_width] = color.unsqueeze(1).unsqueeze(2)
            else:
                img[:, 112:112+height, pos:pos+bar_width] = color.unsqueeze(1).unsqueeze(2)

        # Add a horizontal line at y=0
        img[:, 111:113, :] = torch.tensor([0, 0, 0]).unsqueeze(1).unsqueeze(2).expand(3, 2, total_width)  # Black line

        img = img.unsqueeze(0)
        img = F.interpolate(img, size=(224, 224), mode='bilinear', align_corners=False)
        img = img.squeeze(0)

    return img


def gen_colors(n, seed=37):
    np.random.seed(seed)
    
    def generate_colors(count):
        # Generate colors as integers between 0 and 255
        return np.random.randint(0, 256, size=(count, 3))
    
    # Generate initial set of colors
    colors_int = generate_colors(n)
    
    # Check for uniqueness and generate more if needed
    unique_colors_int = np.unique(colors_int, axis=0)
    while len(unique_colors_int) < n:
        additional_colors = generate_colors(n - len(unique_colors_int))
        colors_int = np.vstack((unique_colors_int, additional_colors))
        unique_colors_int = np.unique(colors_int, axis=0)
    
    # Convert to float and round to 3 decimal places
    unique_colors_float = np.round(unique_colors_int.astype(float) / 255, 3)
    
    return unique_colors_float[:n]


"""
haphazard.data.datasets.dry_bean
--------------------------------
Dry Bean dataset implementation for multi-class classification.
"""

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.io import arff

from ...base_dataset import BaseDataset
from ...datasets import register_dataset
from ....utils.file_utils import find_file


@register_dataset("dry_bean")
class DryBean(BaseDataset):
    """
    Dry Bean dataset for multi-class classification.

    Attributes
    ----------
    name : str
        Name of the dataset.
    haphazard_type : str
        Indicates dataset type, e.g., "controlled".
    task : str
        Task type, e.g., "classification".
    num_classes : int
        Number of unique output classes.

    Notes
    -----
    - Class labels are mapped from strings to integers.
    - Original dataset has ordered labels as follows:
        Index         -> Label
        0 to 2026     -> 0
        2027 to 3348  -> 1
        3349 to 3870  -> 2
        3871 to 5500  -> 3
        5501 to 7428  -> 4
        7429 to 10064 -> 5
        10065 to 13610 -> 6
    - Data is shuffled for randomized experiments.
    """

    def __init__(self, base_path: str = "./", **kwargs) -> None:
        """
        Initialize the Dry Bean dataset instance.

        Parameters
        ----------
        base_path : str, default="./"
            Base directory under which the dataset file is located.
        **kwargs
            Additional keyword arguments passed to the parent constructor.
        """
        self.name = "dry_bean"
        self.haphazard_type = "controlled"
        self.task = "classification"
        self.num_classes = 7

        super().__init__(base_path=base_path, **kwargs)

    def read_data(self, base_path: str = "./") -> tuple[NDArray[np.float64], NDArray[np.int64]]:
        """
        Load and process the Dry Bean dataset.

        Parameters
        ----------
        base_path : str, default="./"
            Directory under which the dataset file is searched.

        Returns
        -------
        tuple[NDArray[np.float64], NDArray[np.int64]]
            x : ndarray of shape (n_samples, n_features)
                Feature matrix with dtype float64.
            y : ndarray of shape (n_samples,)
                Integer label vector with dtype int64.
        """
        data_path = find_file(base_path, "Dry_Bean_Dataset.arff")
        data, _ = arff.loadarff(data_path)
        df = pd.DataFrame(data)

        # Decode string labels and map to integers
        df["Class"] = df["Class"].str.decode("utf-8")
        class_encoding = {
            "SEKER": 0,
            "BARBUNYA": 1,
            "BOMBAY": 2,
            "CALI": 3,
            "HOROZ": 4,
            "SIRA": 5,
            "DERMASON": 6,
        }
        df["Class"] = df["Class"].map(class_encoding)

        # Shuffle rows for randomized experiments
        df = df.sample(frac=1.0, random_state=42)

        x = df.drop(columns=["Class"]).to_numpy(dtype=np.float64)
        y = df["Class"].to_numpy(dtype=np.int64)

        return x, y


__all__ = ["DryBean"]



# Code from https://github.com/Rohit102497/packetLSTM_HaphazardInputs/blob/main/Baselines/Models/OLIFL.py
# Code from https://github.com/youdianlong/OLIFL

import numpy as np
import random,time
import copy,random
from haphazard.utils import seed_everything
from sklearn import preprocessing
from tqdm.auto import tqdm
import math

# np.warnings.filterwarnings('error', category=np.VisibleDeprecationWarning)  # Depricated
import warnings
warnings.filterwarnings('error', category=np.exceptions.VisibleDeprecationWarning)       

def findCommonKeys(classifier, row):  # find the common keys of two dictionaries
    return (set(classifier.keys()) & set(row.keys()))


def findDifferentKeys(dict1, dict2):
    return (set(dict1.keys()) - set(dict2.keys()))


def subsetDictionary(dictionary, intersection):  # extract subset of key-val pairs if in
    return dict((value, dictionary[value]) for value in intersection)


def dict2NumpyArray(dictionary):
    return np.array([dictionary[key] for key in sorted(dictionary.keys())])


def dotDict(dict1, dict2):
    returnValue = 0
    for key in dict1:
        returnValue += dict1[key] * dict2[key]
    return returnValue


def NumpyArray2Dict(numpyArray, keys):
    return {k: v for k, v in zip(keys, numpyArray)}

class OLIFL:
    def __init__(self, C, option,seed):
        self.C = C
        self.option = option
        self.seed = seed
        seed_everything(self.seed)

        self.rounds = 1

    def custom_set_data(self,X, Y, X_haphazard, mask):
        self.X = []
        for t in range(len(Y)):
            dct = {}
            for f in range(X.shape[1]):
                if mask[t][f]:
                    dct[f]=X_haphazard[t][f]
            self.X.append(dct)
        self.y = Y
    def parameter_set(self, X,loss):
        inner_product=dotDict(X, X)
        if inner_product == 0:
            inner_product = 1

        if self.option == 0: return loss / inner_product
        if self.option == 1: return np.minimum(self.C, 2*loss / inner_product)

    def set_classifier(self):
        self.weights = {key: 0 for key in self.X[0].keys()}
        self.u_count={key: 1 for key in self.X[0].keys()}
        self.stability = self.X[0]
        self.count = dict()
        self.A_ =  dict()
        self.A  =  dict()
        self.keyCount=dict()
        self.n_num = 0
        self.e_num = 0
        self.s_num = 0
        self.sum_loss=0
    def update_stability(self,X):
        e_stability=self.stability
        for key in X.keys():
            if key not in self.count.keys():
                self.count[key] = 1
                self.keyCount[key]=1
                self.A_[key]=X[key]
                self.A[key]=X[key]
                self.stability[key]=0.000001
            else:

                self.count[key] +=1
                self.A_[key]=self.A[key]
                self.A[key]=self.A[key] + (X[key] - self.A[key]) / self.count[key]
                self.stability[key]=(self.count[key]-1)/self.count[key]**2*(X[key]-self.A_[key])**2+(self.count[key]-1)/self.count[key]*self.stability[key]
        return e_stability
    def upKeyCount(self,X): 
        sum1 = 0
        e_KeyCount=self.keyCount
        for key in X.keys():
            sum1 += self.stability[key]
        for key in X.keys():
            self.keyCount[key] = self.stability[key] / sum1
        return e_KeyCount

    def predict(self,X):
        y_pre=np.sum(np.array([X[k] * self.keyCount[k] * self.weights[k] for k in X.keys()],dtype=object))
        return y_pre
    



    def expand_space(self,X):
        self.n_keys=0
        self.e_keys=0
        self.s_keys=0
        e_weights=self.weights
        for key in findDifferentKeys(X, self.weights):
            self.weights[key] = 0
            self.u_count[key] = 1
            self.n_keys+=1
        for key in findDifferentKeys(self.weights,X):
            X[key] = 0
            self.e_keys+=1
        for key in findCommonKeys(X,self.weights):
            self.u_count[key] += 1
            self.s_keys=+1
        return X,e_weights

    def fit(self,X, Y, X_haphazard, mask):#
        self.custom_set_data(X, Y, X_haphazard, mask)
        for i in range(self.rounds):
            start = time.time()
            Y_pred=[] 
            Y_logits=[]
            
            train_error, train_loss = 0, 0
            train_error_vector, train_loss_vector, train_acc_vector = [], [], []

            TN = 0
            self.set_classifier()

            l=0
            ###########Change dataformat to required dictionary format#############
            for t in tqdm(range(0, len(self.y))):
                l+=1
                row, _ =self.expand_space(self.X[t])
                if len(row) == 0:
                    train_error_vector.append(train_error / (l + 1))
                    train_loss_vector.append(train_loss / (l + 1))
                    train_acc_vector.append(1 - train_error / (l + 1))
                    continue
                sta_ = self.update_stability(row)
                KC_ = self.upKeyCount(row)
                y_value = self.predict(row)
                y_hat = np.sign(y_value)

                y_hat_logit = 1/(1+np.exp(y_hat))
                Y_logits.append(y_hat_logit)
                y_up = 2*self.y[t]-1

                if y_hat != y_up:
                    train_error += 1
                    if self.y[i] == 1:
                        TN += 1
                loss = (np.maximum(0, (1 - (y_up) * y_value)))
                tao = self.parameter_set(row, loss)
                self.weights=self.upweight(tao,row,y_up)
                train_error_vector.append(train_error / (l + 1))
                train_loss += loss
                train_loss_vector.append(train_loss / (l + 1))
                train_acc_vector.append(1 - (train_error / (l + 1)))
                Y_pred.append([0 if y_hat==-1.0 else 1])
            x = type(Y_logits[0])
            for y in range(len(Y_logits)):
                if x!=type(Y_logits[y]):
                    Y_logits[y] = Y_logits[y][0]
            return Y_pred, Y_logits

    def upweight(self,tao,X,y):


        return {key: self.weights[key] + tao * (y) *X[key]*self.keyCount[key] for key in X.keys()}


import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error
from sklearn import svm
from .ftrl_adp import *

def svm_classifier(train_x, train_y,test_x, test_y):
    best_score=0
    best_C=-1
    for C in [0.001, 0.01, 0.1, 1, 10, 100]:
        clf = svm.LinearSVC(C=C,max_iter=100000)
        clf.fit(train_x,train_y)
        score = clf.score(test_x, test_y)
        if score>best_score:
            best_score=score
            best_C=C
    return  best_score,best_C

def calculate_svm_error(X_input, Y_label,n):
    length=int(0.7*n)
    X_train = X_input[:length, :]
    Y_train = Y_label[:length]
    X_test = X_input[length:, :]
    Y_test = Y_label[length:]
    best_score, best_C = svm_classifier(X_train, Y_train, X_test, Y_test)
    error = 1.0 - best_score
    return error,best_C

def generate_X(n,X_input,Y_label,decay_choice,contribute_error_rate):
    errors=[]
    decays=[]
    predict=[]
    mse=[]

    classifier=FTRL_ADP(decay = 1.0, L1=0., L2=0., LP = 1., adaptive =True,n_inputs = X_input.shape[1])
    for row in range(n):
        indices = [i for i in range(X_input.shape[1])]
        x = X_input[row]
        y = Y_label[row]
        p, decay,loss,w = classifier.fit(indices, x, y ,decay_choice,contribute_error_rate)
        error = [int(np.abs(y - p) > 0.5)]

        errors.append(error)
        decays.append(decay)
        predict.append(p)
        mse.append(mean_squared_error(predict[:row+1], Y_label[:row+1]))

    X_Zero_CER = np.cumsum(errors) / (np.arange(len(errors)) + 1.0)
    svm_error,_ =calculate_svm_error(X_input[:,1:], Y_label, n)

    return X_Zero_CER, svm_error

def generate_cap(n,X_input,Y_label,decay_choice,contribute_error_rate):
    errors=[]
    decays=[]

    classifier=FTRL_ADP(decay = 1.0, L1=0., L2=0., LP = 1., adaptive =True,n_inputs = X_input.shape[1])
    for row in range(n):
        indices = [i for i in range(X_input.shape[1])]
        x = X_input[row].data
        y = Y_label[row]
        p, decay,loss,w= classifier.fit(indices, x, y ,decay_choice,contribute_error_rate)
        error = [int(np.abs(y - p) > 0.5)]

        errors.append(error)
        decays.append(decay)

    Z_imp_CER = np.cumsum(errors) / (np.arange(len(errors)) + 1.0)

    return Z_imp_CER
def generate_tra(n, X_input, Y_label, decay_choice, contribute_error_rate):
    errors = []
    classifier = FTRL_ADP(decay=1.0, L1=0., L2=0., LP=1., adaptive=True, n_inputs=len(X_input[-1]))
    for row in range(n):
        indices = [i for i in range(len(X_input[row]))]
        x = np.array(X_input[row]).data
        y = Y_label[row]
        p, decay, loss ,w= classifier.fit(indices, x, y, decay_choice, contribute_error_rate)
        error = [int(np.abs(y - p) > 0.5)]
        errors.append(error)
    imp_CER=np.cumsum(errors) / (np.arange(len(errors)) + 1.0)

    return imp_CER

def draw_cap_error_picture(Z_imp_CER,X_Zero_CER,svm_error):
    n=len(Z_imp_CER)
    plt.figure(figsize=(6, 4))
    plt.ylim((0, 1))
    plt.xlim((0, n))
    plt.ylabel("CER")  #
    x = range(n)

    plt.plot(x, Z_imp_CER  , color='green')    # the error of z_imp
    plt.plot(x, X_Zero_CER, color='blue')     # the error of x_zero
    plt.plot(x, [svm_error] * n ,color='red') # the error of svm

    plt.title("The error of z_imp_CER,x_zero_CER,SVM_CER")
    plt.show()
    plt.clf()

def draw_tra_error_picture(error_arr_Z,error_arr_X):
    n=len(error_arr_Z)
    plt.figure(figsize=(10, 8))
    plt.ylim((0, 1.0))
    plt.xlim((0, n))
    plt.ylabel("CER")  #

    x = range(n)
    plt.plot(x, error_arr_Z, color='green')   # z_imp error picture

    plt.plot(x, error_arr_X, color='blue')    # x_imp error picture

    plt.title("The CER of trapezoid data stream")
    plt.show()
    plt.clf()


"""
haphazard.models.model_zoo.olvf
-------------------------------
Wrapper over OLVF model for binary and multi-class classification.

Implements `RunOLVF` runner class and a `MultiClassWrapper` for handling
multi-class tasks using a One-vs-Rest strategy.
"""

import time
import copy
from typing import Any

import numpy as np
from numpy.typing import NDArray
from tqdm.auto import tqdm

from .olvf import OLVF
from ...base_model import BaseModel, BaseDataset
from ...model_zoo import register_model
from ....utils.seeding import seed_everything


class MultiClassWrapper:
    """
    Wraps a binary OLVF model to perform multi-class classification using
    the One-vs-Rest (OvR) strategy.

    Each class is handled by a separate copy of the base binary model.

    Attributes
    ----------
    num_classes : int
        Number of unique classes in the dataset.
    models : list[OLVF]
        Independent OLVF instances for each class.
    """

    def __init__(self, model_instance: OLVF, num_classes: int) -> None:
        """
        Initialize the One-vs-Rest wrapper.

        Parameters
        ----------
        model_instance : OLVF
            A fully initialized binary OLVF model instance.
        num_classes : int
            Total number of output classes.
        """
        self.num_classes: int = num_classes
        self.models: list[OLVF] = [copy.deepcopy(model_instance) for _ in range(num_classes)]

    def partial_fit(
        self,
        X: NDArray[np.float64],
        X_mask: NDArray[np.bool_],
        y_true: int,
    ) -> tuple[int, list[float]]:
        """
        Perform a single online update step for one instance.

        Parameters
        ----------
        X : NDArray[np.float64]
            Input feature vector, shape (n_features,).
        X_mask : NDArray[np.bool_]
            Binary mask for available features.
        y_true : int
            Ground-truth label in [0, num_classes).

        Returns
        -------
        tuple[int, list[float]]
            - Predicted class index.
            - Logits for each class, length `num_classes`.
        """
        logits = [0.0 for _ in range(self.num_classes)]

        for cls_idx, model in enumerate(self.models):
            binary_label = 1 if y_true == cls_idx else 0
            _, logit = model.partial_fit(X, X_mask, binary_label)
            logits[cls_idx] = float(logit)

        y_pred = int(np.argmax(logits))
        return y_pred, logits


@register_model("olvf")
class RunOLVF(BaseModel):
    """
    Runner class for OLVF model.

    Attributes
    ----------
    name : str
        Model name.
    tasks : set[str]
        Supported task types.
    deterministic : bool
        Whether model is deterministic.
    hyperparameters : set[str]
        Supported hyperparameters.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initialize the OLVF runner class.

        Parameters
        ----------
        **kwargs
            Optional parameters forwarded to `BaseModel`.
        """
        self.name = "OLVF"
        self.tasks = {"classification"}
        self.deterministic = True
        self.hyperparameters = {"C", "C_bar", "B", "reg"}

        super().__init__(**kwargs)

    def fit(
        self,
        dataset: BaseDataset,
        model_params: dict[str, Any] | None = None,
        seed: int = 42,
    ) -> dict[str, Any]:
        """
        Run the OLVF model on a dataset.

        Parameters
        ----------
        dataset : BaseDataset
            Dataset object with `.load_data()` and `.load_mask()`.
        model_params : dict[str, Any] | None, optional
            Hyperparameters for OLVF model.
        seed : int, default=42
            Random seed for reproducibility.

        Returns
        -------
        dict[str, Any]
            Dictionary containing:
            - labels: Ground truth labels.
            - preds: Predicted labels.
            - logits: Model output logits or probabilities.
            - time_taken: Time taken for full dataset pass.
            - is_logit: Indicates whether scores are logits (True) or probabilities (False).

        Note
        ----
        The dataset object now provides normalized, masked samples via __iter__ and __getitem__,
        enabling online (per-instance) learning without full-batch loading.
        Slicing and iteration behavior is controlled by BaseDataset.
        """
        # --- Validate task ---
        if dataset.task not in self.tasks:
            raise ValueError(
                f"Model '{self.__class__.__name__}' does not support '{dataset.task}'. "
                f"Supported task(s): {self.tasks}"
            )
        
        model_params = model_params or {}

        # Set random seed and initialize base_model
        seed_everything(seed)
        base_model: OLVF = OLVF(**model_params, n_feat0=dataset.n_features)

        if dataset.task == "regression":
            raise NotImplementedError("Regression task not supported for OLVF.")

        elif dataset.task == "classification":
            if dataset.num_classes is None:
                raise ValueError(f"'{dataset.name}.num_classes' cannot be None for classification task.")
            
            if dataset.num_classes == 2:
                model: OLVF | MultiClassWrapper = base_model
            else:
                model = MultiClassWrapper(base_model, num_classes=dataset.num_classes)

            pred_list: list[int | float] = []
            logit_list: list[list[float] | float] = []

            # --- Train model ---
            start_time = time.perf_counter()

            for x, mask, y in tqdm(
                dataset,
                total=dataset.n_samples,
                desc="Running OLVF",
            ):
                pred, logit = model.partial_fit(x, mask, int(y))
                pred_list.append(pred)
                logit_list.append(logit)

            end_time = time.perf_counter()
            time_taken = end_time - start_time

            # --- Final formatting ---
            labels = np.asarray(dataset.y, dtype=np.int64)
            preds = np.asarray(pred_list, dtype=np.int64)
            logits = np.asarray(logit_list, dtype=np.float64)

            # --- Sanity checks ---
            if dataset.num_classes == 2:
                assert logits.ndim == 1, (
                    f"Expected logits to be 1D for binary classification, got {logits.shape}."
                )
            else:
                assert logits.ndim == 2, (
                    f"Expected logits to be 2D for multi-class classification, got {logits.shape}."
                )

            is_logit = True  # OLVF model returns logits

            return {
                "labels": labels,
                "preds": preds,
                "logits": logits,
                "time_taken": time_taken,
                "is_logit": is_logit,
            }

        # Fallback for unsupported task
        raise ValueError(f"Unknown task type: '{dataset.task}'")


__all__ = ["RunOLVF"]



from transforms.trapezoidal_transform_function2 import TrapezoidalTransformFunction2
import numpy as np
from concurrent.futures import ProcessPoolExecutor
from em.expectation_maximization import ExpectationMaximization
from em.embody import _em_step_body_, _em_step_body, _em_step_body_row

class TrapezoidalExpectationMaximization2(ExpectationMaximization):
    def __init__(self, cont_indices, ord_indices, window_size,window_width, sigma_init=None):
        self.transform_function = TrapezoidalTransformFunction2(cont_indices, ord_indices, window_size=window_size,window_width=window_width)
        self.cont_indices = cont_indices
        self.ord_indices = ord_indices
        # we assume boolean array of indices
        p = len(cont_indices)
        # By default, sigma corresponds to the correlation matrix of the permuted dataset (ordinals appear first, then continuous)
        if sigma_init is not None:
            self.sigma = sigma_init
        else:
            self.sigma = np.identity(p)
        # track what iteration the algorithm is on for use in weighting samples
        self.iteration = 1

    def partial_fit_and_predict(self, X_batch,cont_indices,ord_indices, max_workers=4, num_ord_updates=2, decay_coef=0.5, sigma_update=True, marginal_update = True, sigma_out=False):
        """
        Updates the fit of the copula using the data in X_batch and returns the 
        imputed values and the new correlation for the copula

        Args:
            X_batch (matrix): data matrix with entries to use to update copula and be imputed
            max_workers (positive int): the maximum number of workers for parallelism 
            num_ord_updates (positive int): the number of times to re-estimate the latent ordinals per batch
            decay_coef (float in (0,1)): tunes how much to weight new covariance estimates
        Returns:
            X_imp (matrix): X_batch with missing values imputed
        """
        self.cont_indices = cont_indices
        self.ord_indices = ord_indices
        #if not update:
            #old_window = self.transform_function.window
            #old_update_pos = self.transform_function.update_pos
        if marginal_update:
            self.transform_function.partial_fit(X_batch,cont_indices,ord_indices)
        # update marginals with the new batch
        #self.transform_function.partial_fit(X_batch)
        res = self._fit_covariance(X_batch, max_workers, num_ord_updates, decay_coef, sigma_update, sigma_out)
        if sigma_out:
            Z_batch_imp, sigma = res
        else:
            Z_batch_imp = res
        # Rearrange Z_imp so that it's columns correspond to the columns of X
        Z_imp_rearranged = np.empty(X_batch.shape)
        Z_imp_rearranged[:,self.ord_indices] = Z_batch_imp[:,:np.sum(self.ord_indices)]
        Z_imp_rearranged[:,self.cont_indices] = Z_batch_imp[:,np.sum(self.ord_indices):]
        X_imp = np.empty(X_batch.shape)
        X_imp[:,self.cont_indices] = self.transform_function.partial_evaluate_cont_observed(Z_imp_rearranged, X_batch)
        X_imp[:,self.ord_indices] = self.transform_function.partial_evaluate_ord_observed(Z_imp_rearranged, X_batch)
        #if not update:
            #self.transform_function.window = old_window
            #self.transform_function.update_pos = old_update_pos
         #pass
        if sigma_out:
            return Z_imp_rearranged,X_imp, sigma
        else:
            return Z_imp_rearranged,X_imp

    def _fit_covariance(self, X_batch, max_workers=4, num_ord_updates=2, decay_coef=0.5, update=True, sigma_out=False, seed=1):
        """
        Updates the covariance matrix of the gaussian copula using the data 
        in X_batch and returns the imputed latent values corresponding to 
        entries of X_batch and the new sigma

        Args:
            X_batch (matrix): data matrix with which to update copula and with entries to be imputed
            max_workers: the maximum number of workers for parallelism 
            num_ord_updates: the number of times to restimate the latent ordinals per batch
            decay_coef (float in (0,1)): tunes how much to weight new covariance estimates
        Returns:
            sigma (matrix): an updated estimate of the covariance of the copula
            Z_imp (matrix): estimates of latent values in X_batch
        """
        Z_ord_lower, Z_ord_upper = self.transform_function.partial_evaluate_ord_latent(X_batch)
        Z_ord = self._init_Z_ord(Z_ord_lower, Z_ord_upper, seed)
        Z_cont = self.transform_function.partial_evaluate_cont_latent(X_batch)
        # Latent variable matrix with columns sorted as ordinal, continuous
        Z = np.concatenate((Z_ord, Z_cont), axis=1)
        batch_size, p = Z.shape
        # track previous sigma for the purpose of early stopping
        prev_sigma = self.sigma
        s=prev_sigma.shape[0]
        if p>s:
            temp=np.zeros([s,p-s])
            prev_sigma=np.insert(prev_sigma,[s],temp,axis=1)
            temp=np.zeros([p-s,p])
            prev_sigma=np.insert(prev_sigma,[s],temp,axis=0)
            for i in range(s,p):
                prev_sigma[i][i]=1.0

        Z_imp = np.zeros((batch_size, p))
        C = np.zeros((p, p))
        if max_workers==1:
            try:
                C, Z_imp, Z = _em_step_body(Z,Z_ord_lower, Z_ord_upper, prev_sigma, num_ord_updates)
            except:
                np.savetxt("X_batch.txt",X_batch)
                np.savetxt("z_batch.txt",Z)
                np.savetxt("prev_sigma.txt",prev_sigma)
        else:

            divide = batch_size/max_workers * np.arange(max_workers+1)
            divide = divide.astype(int)
            args = [(Z[divide[i]:divide[i+1],:],Z_ord_lower[divide[i]:divide[i+1],:], Z_ord_upper[divide[i]:divide[i+1],:], prev_sigma, num_ord_updates) for i in range(max_workers)]
            # divide each batch into max_workers parts instead of n parts
            with ProcessPoolExecutor(max_workers=max_workers) as pool:
                res = pool.map(_em_step_body_, args)
                for i,(C_divide, Z_imp_divide, Z_divide) in enumerate(res):
                    Z_imp[divide[i]:divide[i+1],:] = Z_imp_divide
                    Z[divide[i]:divide[i+1],:] = Z_divide # not necessary if we only do on EM iteration
                    C += C_divide
        C = C/batch_size
        sigma = np.cov(Z_imp, rowvar=False) + C
        L=sigma.shape[0]
        for i in range(L):
            if sigma[i][i]==0:sigma[i][i]=1.0
        sigma = self._project_to_correlation(sigma)
        if update:
            self.sigma = sigma*decay_coef + (1 - decay_coef)*prev_sigma
            prev_sigma = self.sigma
            self.iteration += 1
        if sigma_out:
            if update:
                sigma = self.get_sigma()
            else:
                sigma = self.get_sigma(sigma*decay_coef + (1 - decay_coef)*prev_sigma)
            return Z_imp, sigma
        else:
            return Z_imp

    def get_sigma(self, sigma=None):
        if sigma is None:
            sigma = self.sigma
        sigma_rearranged = np.empty(sigma.shape)
        sigma_rearranged[np.ix_(self.ord_indices,self.ord_indices)] = sigma[:np.sum(self.ord_indices),:np.sum(self.ord_indices)]
        sigma_rearranged[np.ix_(self.cont_indices,self.cont_indices)] = sigma[np.sum(self.ord_indices):,np.sum(self.ord_indices):]
        sigma_rearranged[np.ix_(self.cont_indices,self.ord_indices)] = sigma[np.sum(self.ord_indices):,:np.sum(self.ord_indices)]
        sigma_rearranged[np.ix_(self.ord_indices,self.cont_indices)] =  sigma_rearranged[np.ix_(self.cont_indices,self.ord_indices)].T
        return sigma_rearranged

    def _init_sigma(self, sigma):
        sigma_new = np.empty(sigma.shape)
        sigma_new[:np.sum(self.ord_indices),:np.sum(self.ord_indices)] = sigma[np.ix_(self.ord_indices,self.ord_indices)]
        sigma_new[np.sum(self.ord_indices):,np.sum(self.ord_indices):] = sigma[np.ix_(self.cont_indices,self.cont_indices)]
        sigma_new[np.sum(self.ord_indices):,:np.sum(self.ord_indices)] = sigma[np.ix_(self.cont_indices,self.ord_indices)] 
        sigma_new[:np.sum(self.ord_indices),np.sum(self.ord_indices):] = sigma[np.ix_(self.ord_indices,self.cont_indices)] 
        self.sigma = sigma_new

    def change_point_test(self, x_batch, decay_coef, nsample=100, max_workers=4):
        n,p = x_batch.shape
        #xsample = np.random.multivariate_normal(np.zeros(p), self.sigma, (nsample,n))
        statistics = np.zeros((nsample,3))
        sigma_old = self.get_sigma()
        _, sigma_new = self.partial_fit_and_predict(x_batch, decay_coef=decay_coef, max_workers=max_workers, marginal_update=True, sigma_update=False, sigma_out=True)
        s = self.get_matrix_diff(sigma_old, sigma_new)
        # generate incomplete mixed data samples
        for i in range(nsample):
            np.random.seed(i)
            z = np.random.multivariate_normal(np.zeros(p), sigma_old, n)
            # mask
            x = np.empty(x_batch.shape)
            x[:,self.cont_indices] = self.transform_function.partial_evaluate_cont_observed(z)
            x[:,self.ord_indices] = self.transform_function.partial_evaluate_ord_observed(z)
            loc = np.isnan(x_batch)
            x[loc] = np.nan
            _, sigma = self.partial_fit_and_predict(x, decay_coef=decay_coef, max_workers=max_workers, marginal_update=False, sigma_update=False, sigma_out=True)
            statistics[i,:] = self.get_matrix_diff(sigma_old, sigma)
        # compute test statistics
        pval = np.zeros(3)
        for j in range(3):
            pval[j] = np.sum(s[j]<statistics[:,j])/(nsample+1)
        self._init_sigma(sigma_new)
        return pval, s

        # compute test statistics
    def get_matrix_diff(self, sigma_old, sigma_new, type = 'F'):
        '''
        Return the correlation change tracking statistics, as some matrix norm of normalized matrix difference.
        Support three norms currently: 'F' for Frobenius norm, 'S' for spectral norm and 'N' for nuclear norm. User-defined norm can also be used.
        '''
        p = sigma_old.shape[0]
        u, s, vh = np.linalg.svd(sigma_old)
        factor = (u * np.sqrt(1/s) ) @ vh
        diff = factor @ sigma_new @ factor
        if type == 'F':
            return np.linalg.norm(diff-np.identity(p))
        else:
            _, s, _ = np.linalg.svd(diff)
            if type == 'S':
                return max(abs(s-1))
            if type == 'N':
                return np.sum(abs(s-1))


import warnings
warnings.filterwarnings("ignore")
import sys
sys.path.append("..")
from em.online_expectation_maximization import OnlineExpectationMaximization
from evaluation.helpers import *
from onlinelearning.online_learning import *
from onlinelearning.ensemble import *
import math

if __name__ == '__main__':
    dataset = "wdbc" # australian,ionosphere,german,diabetes,wdbc,credit,ionosphere

    #getting  hyperparameter
    contribute_error_rate, window_size_denominator, batch_size_denominator, decay_coef_change,decay_choice,shuffle =get_cap_hyperparameter(dataset)
    MASK_NUM=1
    X = pd.read_csv("../dataset/MaskData/"+dataset+"/X_process.txt",sep=" " ,header=None)
    if dataset == "a8a":
        X = abs(X)
    Y_label = pd.read_csv("../dataset/Datalabel/" + dataset + "/Y_label.txt", sep=' ', header=None)
    X_masked=mask_types(X,MASK_NUM,seed=1) #arbitrary setting Nan
    X = X.values
    Y_label = Y_label.values

    all_cont_indices=get_cont_indices(X_masked)
    all_ord_indices=~all_cont_indices

    n = X_masked.shape[0]
    feat = X_masked.shape[1]
    Y_label = Y_label.flatten()

    #setting hyperparameter
    max_iter = n
    BATCH_SIZE = 1
    WINDOW_SIZE = min(500, int(n/10)) # Arijit: Window size is basically buffer size
    # max_iter = batch_size_denominator * 2
    # BATCH_SIZE = math.ceil(n / batch_size_denominator)
    # WINDOW_SIZE = math.ceil(n / window_size_denominator) # Arijit: Window size is basically buffer size
    NUM_ORD_UPDATES = 1
    batch_c = 8

    # start online copula imputation
    oem = OnlineExpectationMaximization(all_cont_indices, all_ord_indices, window_size=WINDOW_SIZE)
    j = 0
    X_imp = np.empty(X_masked.shape)
    Z_imp = np.empty(X_masked.shape)
    X_masked=np.array(X_masked)
    while j <= max_iter:
        start = (j * BATCH_SIZE) % n
        end = ((j + 1) * BATCH_SIZE) % n
        if end < start:
            indices = np.concatenate((np.arange(end), np.arange(start, n, 1)))
        else:
            indices = np.arange(start, end, 1)
        if decay_coef_change == 1:
            this_decay_coef = batch_c / (j/(batch_size_denominator*2) + batch_c)
            # this_decay_coef = batch_c / (j + batch_c)
        else:
            this_decay_coef = 0.5
        Z_imp[indices, :], X_imp[indices, :] = oem.partial_fit_and_predict(X_masked[indices, :], max_workers=1,decay_coef=this_decay_coef)
        j += 1

    # get the error of FOBOS and SVM
    temp = np.ones((n, 1))

    X_masked = pd.DataFrame(X_masked)
    X_zeros = X_masked.fillna(value=0)
    X_input_zero = np.hstack((temp, X_zeros))
    if shuffle == True:
        perm = np.arange(n)
        np.random.seed(1)
        np.random.shuffle(perm)
        Y_label = Y_label[perm]
        X_input_zero = X_input_zero[perm]
    X_Zero_CER, svm_error = generate_X(n, X_input_zero, Y_label, decay_choice, contribute_error_rate)

    #get the error of latent space
    temp_zim = np.ones((n, 1))
    X_input_z_imp = np.hstack((temp, Z_imp))
    if shuffle == True:
        perm = np.arange(n)
        np.random.seed(1)
        np.random.shuffle(perm)
        X_input_z_imp = X_input_z_imp[perm]

    # Z_imp_CER = generate_cap(n,X_input_z_imp,Y_label,decay_choice,contribute_error_rate)
    Z_imp_CER = ensemble(n,X_input_z_imp,X_input_zero,Y_label,decay_choice,contribute_error_rate)

    draw_cap_error_picture(Z_imp_CER,X_Zero_CER,svm_error)



"""
haphazard.normalization.normalizer_zoo.unit_vector
--------------------------------------------------
UnitVectorNormalization — L2 normalization to unit length.

This module implements normalization that rescales observed features to
form a unit vector, ensuring the L2 norm equals 1. It is a stateless
normalization method, applied directly on the input without tracking
running statistics.

Useful for models sensitive to feature magnitude (e.g., cosine similarity).
"""

from numpy.typing import NDArray
import numpy as np

from ...base_normalizer import BaseNormalizer
from ...normalizer_zoo import register_normalizer


# ----------------------------------------------------------------------
# Unit Vector Normalization
# ----------------------------------------------------------------------
@register_normalizer("unit_vector")
class UnitVectorNormalization(BaseNormalizer):
    """
    Normalizes observed features to form a unit vector (L2 norm = 1).

    This method scales observed feature values so that their Euclidean norm
    equals 1. It performs no parameter updates and is fully deterministic.

    Notes
    -----
    - No running statistics are maintained.
    """

    def update_params(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> None:
        """No parameter updates required for unit vector normalization."""
        pass

    def normalize(self, x: NDArray[np.float64], indices: NDArray[np.int64]) -> NDArray[np.float64]:
        """Normalize observed features to have unit L2 norm."""
        vec = x[indices]
        norm = np.linalg.norm(vec)
        if norm > 0:
            x[indices] = vec / norm
        return x



"""
haphazard.data.datasets
-----------------------
Dataset registry and dynamic loader.

This module automatically imports all dataset definitions in the current
directory and provides a convenient `load_dataset` interface.
"""

import importlib
import pkgutil
from typing import Type, Dict, Any, Callable

from ..base_dataset import BaseDataset, MaskScheme


# -------------------------------------------------------------------------
# Dataset registry
# -------------------------------------------------------------------------
_DATASET_REGISTRY: Dict[str, Type[BaseDataset]] = {}


def register_dataset(name: str) -> Callable[[Type[BaseDataset]], Type[BaseDataset]]:
    """
    Decorator to register a dataset class in the global registry.

    Parameters
    ----------
    name : str
        The name used to register the dataset (case-insensitive).

    Returns
    -------
    Callable[[Type[BaseDataset]], Type[BaseDataset]]
        A decorator that registers the dataset class in the registry.

    Raises
    ------
    TypeError
        If the decorated class is not a subclass of `BaseDataset`.
    ValueError
        If a dataset with the same name is already registered.

    Examples
    --------
    >>> @register_dataset("my_data")
    ... class MyDataset(BaseDataset):
    ...     pass
    """
    def decorator(cls: Type[BaseDataset]) -> Type[BaseDataset]:
        if not issubclass(cls, BaseDataset):
            raise TypeError(f"Cannot register '{cls.__name__}': not a subclass of BaseDataset.")
        
        key = name.lower()
        if key in _DATASET_REGISTRY:
            raise ValueError(
                f"Duplicate dataset name detected: '{key}'. "
                "(Note: The 'name' used to register a dataset is case-insensitive)"
            )
        
        _DATASET_REGISTRY[key] = cls
        return cls

    return decorator


# -------------------------------------------------------------------------
# Dynamic import of all submodules
# -------------------------------------------------------------------------
# Ensures that all decorated datasets are registered in the global registry.
_package_name = __name__

for module_info in pkgutil.iter_modules(__path__, prefix=f"{_package_name}."):
    importlib.import_module(module_info.name)


# -------------------------------------------------------------------------
# Public loader
# -------------------------------------------------------------------------
def load_dataset(
    name: str,
    base_path: str = "./",
    scheme: MaskScheme | None = None,
    availability_prob: float = 0.5,
    num_chunks: int = 5,
    mask_seed: int = 42,
    norm: str = "none",
    **kwargs: Any,
) -> BaseDataset:
    """
    Load a dataset by name.

    Parameters
    ----------
    name : str
        Name of the dataset to load.
    base_path : str, default="./"
        Path to the directory containing the raw data.
    scheme : MaskScheme | None, default=None
        Masking scheme ("intrinsic", "probabilistic", etc.).
    availability_prob : float, default=0.5
        Probability of feature availability for probabilistic masking.
    num_chunks : int, default=5
        Number of chunks for chunk-based mask schemes.
    mask_seed : int, default=42
        Random seed for mask generation.
    norm : str, default="none"
        Normalization method to apply ("none", "zscore", etc.).
    **kwargs : Any
        Additional arguments passed to the dataset constructor.
    """
    key = name.lower()
    if key not in _DATASET_REGISTRY:
        raise ValueError(
            f"Unknown dataset '{name}'. "
            f"Available datasets: {list(_DATASET_REGISTRY.keys())}"
        )

    dataset_cls = _DATASET_REGISTRY[key]
    return dataset_cls(
        base_path=base_path,
        scheme=scheme,
        availability_prob=availability_prob,
        num_chunks=num_chunks,
        mask_seed=mask_seed,
        norm=norm,
        **kwargs,
    )


# -------------------------------------------------------------------------
# Public exports
# -------------------------------------------------------------------------
__all__ = ["load_dataset", "_DATASET_REGISTRY", "register_dataset"]



"""
haphazard.utils.file_utils
--------------------------
File and path utility functions.

This module provides helper functions for discovering, validating, and
managing files and directories within the project. Useful for
dataset/model loading, configuration discovery, and other filesystem tasks.

Notes
-----
- Functions generally return absolute paths or validated objects.
"""

from pathlib import Path


# -------------------------------------------------------------------------
# File discovery function
# -------------------------------------------------------------------------
def find_file(base_path: str | Path, file_path: str) -> str:
    """
    Recursively search for a file (or relative path) under a base directory.

    Parameters
    ----------
    base_path : str | Path
        Directory under which the file search is performed.
    file_path : str
        File name or relative path to search for.

    Returns
    -------
    : str
        Absolute path of the first matching file found.

    Raises
    ------
    FileNotFoundError
        If the file is not found under the `base_path`.

    Example
    -------
    >>> find_file("/home/user/projects", "config.yaml")
    '/home/user/projects/config.yaml'
    """
    base_path = Path(base_path).resolve()
    for f in base_path.rglob(file_path):
        return str(f)  # Returns the first found instance
    raise FileNotFoundError(f"{file_path} not found under {base_path}")



"""
haphazard.data.datasets.dummy_dataset
-------------------------------------
Implements a dummy dataset for testing and prototyping.
"""

import numpy as np
from numpy.typing import NDArray

from ...base_dataset import BaseDataset, TaskType
from ...datasets import register_dataset


@register_dataset("dummy")
class DummyDataset(BaseDataset):
    """
    Dummy dataset for testing and validation.

    This dataset generates synthetic features and targets for quick
    prototyping, unit testing, or pipeline validation.
    """

    def __init__(
        self,
        base_path: str = "./",
        n_samples: int = 100,
        n_features: int = 10,
        task: TaskType = "classification",
        num_classes: int = 2,
        **kwargs
    ) -> None:
        """
        Initialize the dummy dataset.

        Parameters
        ----------
        base_path : str, default="./"
            Base path to the dataset (unused here).
        n_samples : int, default=100
            Number of samples to generate.
        n_features : int, default=10
            Number of features per sample.
        task : {"classification", "regression"}, default="classification"
            Task type that determines the target generation scheme.
        num_classes : int, default=2
            Number of output classes (used only for classification tasks).
        **kwargs : Any
            Additional arguments passed to :class:`BaseDataset`.
        """
        self.name = "dummy"
        self.haphazard_type = "controlled"
        self.task = task
        self._n_samples = n_samples
        self._n_features = n_features
        self.num_classes = num_classes if task == "classification" else None

        super().__init__(base_path=base_path, **kwargs)

    def read_data(self, base_path: str = "./") -> tuple[NDArray, NDArray]:
        """
        Generate dummy feature and target data.

        Parameters
        ----------
        base_path : str, default="./"
            Base path (unused in this dataset).

        Returns
        -------
        tuple of (NDArray, NDArray)
            Feature matrix `x` of shape (n_samples, n_features)
            and target vector `y` of shape (n_samples,).
        """
        x = np.random.random((self._n_samples, self._n_features))

        if self.task == "classification":
            y = np.random.randint(0, self.num_classes, size=self._n_samples)
        else:
            y = np.random.random((self._n_samples,))

        return x, y


__all__ = ["DummyDataset"]



import numpy as np
import pandas as pd

def DictCompare(W, X, op = 'intersection'):
    if op == 'intersection':
        shared_W_keys = list(W.keys() & X.keys())
        shared_W_keys.sort()
        W = DictToList(W, shared_W_keys)
        X = DictToList(X, shared_W_keys, WorX='X')
        return shared_W_keys, W, X # keys and corresponding values
    if op == 'extraW':
        extra_W_keys = np.sort(np.array(list(W.keys() - X.keys())))
        extra_W_values = DictToList(W, extra_W_keys)
        return extra_W_keys, extra_W_values

def DictToList(dict, indexList = None, WorX = 'weight'):
    df = pd.DataFrame(dict)
    if indexList is not None:
        df_sel = df[indexList]
        if WorX != 'weight':
            return df_sel.values.reshape(1, -1)
        else:
            return df_sel.values.reshape(-1, 1)
    else:
        if WorX != 'weight':
            return df.values.reshape(1, -1)
        else:
            return df.values.reshape(-1, 1)

def MatrixInDict(matrix, dict):
    matrixTemp = matrix.copy()
    '''Always take the full feature space as the dimension of mapped vector'''
    for (r,row) in enumerate(matrix):
        for (c,col) in enumerate(row):
            if dict.get(r) is not None:
                key_new_feature = dict.get(r)
                if key_new_feature is not None:
                    key_new_to_all = key_new_feature.get(c)
                    matrixTemp[r, c] = key_new_to_all
    return matrixTemp


"""
haphazard.normalization
-----------------------
Normalization Subpackage

Provides the `BaseNormalizer` abstract class that all registered normalizers 
must inherit from. Also contains the `normalizer_zoo` subpackage which
maintains all registered normalizers and their loaders.
"""

from .base_normalizer import BaseNormalizer
from .normalizer_zoo import load_normalizer, _NORMALIZER_REGISTRY, register_normalizer


# -------------------------------------------------------------------------
# Public exports
# -------------------------------------------------------------------------

__all__ = [
    "BaseNormalizer",
    "load_normalizer",
    "_NORMALIZER_REGISTRY",
    "register_normalizer"
]



from typing import cast

import torch
import torch.nn as nn

class Embedding(nn.Module):
    '''Create a list of learnable embeddings'''
    def __init__(self, embedding_size: int, init_features: int):
        super(Embedding, self).__init__()

        # Initialize (static) embeddings (no gradient updation)
        self.register_buffer('embedding_list', torch.empty(init_features, embedding_size))
        self.embedding_list = cast(torch.Tensor, self.embedding_list)
        nn.init.kaiming_normal_(self.embedding_list, mode='fan_in', nonlinearity='relu')

        # Initialize learnable embeddings (with gradient updation)
        self.embedding_list_grad = nn.Parameter(self.embedding_list.clone())

        # Indices to be updated in the next step
        self.last_indices = None

    def forward(self, x: torch.Tensor, x_mask: torch.Tensor) -> torch.Tensor:
        indices = torch.where(x_mask)[0]
        
        if self.last_indices is not None:
            with torch.no_grad():
                # Copy the updated values of features present in the last step from learnable to static embeddings
                self.embedding_list[self.last_indices] = self.embedding_list_grad[self.last_indices]

                # Copy the updated static embedding values to learnable embeddings
                self.embedding_list_grad.copy_(self.embedding_list)
        
        # Store indices to be updated in the next step
        self.last_indices = indices

        return torch.einsum('be, b -> be', self.embedding_list_grad[indices], x[indices])
        # tensor                        : shape
        #-----------------------------------------------------------------------
        # self.embedding_list[indices]  : (no. available features x embed size)
        # x[indices]                    : (no. available features)
        # output                        : (no. available features x embed size)


class Aggregation(nn.Module):
    '''Aggregate the feature-embeddings of the available features into instance-embedding'''
    def __init__(self, mode: str):
        super(Aggregation, self).__init__()

        if mode == 'max':
            # Select the max value for each embedding index from available feature embeddings
            self.aggregate = lambda x: torch.max(x, dim=0).values.view(1, -1)

        elif mode == 'min':
            # Select the min value for each embedding index from available feature embeddings
            self.aggregate = lambda x: torch.min(x, dim=0).values.view(1, -1)
        
        elif mode == 'max_magnitude':
            # Select the embedding with largest magnitude
            self.aggregate = lambda x: x[torch.argmax((x**2).sum(dim=1))].view(1, -1)

        elif mode == 'min_magnitude':
            # Select the embedding with smallet magnitude
            self.aggregate = lambda x: x[torch.argmin((x**2).sum(dim=1))].view(1, -1)

        elif mode == 'mean':
            # Take arithmatic mean of all the embeddings
            self.aggregate = lambda x: torch.mean(x, dim=0).view(1, -1)

        elif mode == 'sum':
            # Take arithmatic mean of all the embeddings
            self.aggregate = lambda x: torch.sum(x, dim=0).view(1, -1)

        else:
            raise ValueError(f'`{mode}` is an invalid mode for Aggregation class')
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.aggregate(x)


class LSTMModel(nn.Module):
    '''LSTM classification/regression head implementation'''
    def __init__(self, input_size: int, output_size: int, hidden_size: int, num_layers: int):
        super().__init__()
        self.lstm = nn.LSTM(input_size, hidden_size, num_layers, batch_first=True)
        self.fc = nn.Linear(hidden_size, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        _, (h_n, _) = self.lstm(x)  # Get last hidden state
        x = self.fc(h_n[-1])  # Use last LSTM layer's hidden state
        return x  # Raw logits, activation should be applied externally
    

class TransformerModel(nn.Module):
    '''Transformer classification/regression head implementation'''
    def __init__(self, input_size: int, output_size: int, num_layers: int, num_heads: int, dim_feedforward: int):
        super().__init__()
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=input_size, 
            nhead=num_heads, 
            dim_feedforward=dim_feedforward,
            batch_first=True  
        )
        self.transformer = nn.TransformerEncoder(self.encoder_layer, num_layers=num_layers)
        self.output_layer = nn.Linear(input_size, output_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x.unsqueeze(0)  # Add batch dimension
        x = self.transformer(x)
        x = self.output_layer(x[:, 0])  # Take the first token's output
        return x  # Raw logits, activation should be applied externally


class MLP(nn.Module):
    '''Initialize a MLP by passing hidden-layer-size list'''
    def __init__(self, input_size: int, output_size: int, hidden_layers: list[int], non_linearity: str):
        super(MLP, self).__init__()

        hidden_layers = hidden_layers + [output_size]

        # Choose non-linearity
        non_linear_map = {'relu': nn.ReLU, 'sigmoid': nn.Sigmoid, 'tanh': nn.Tanh}
        self.non_linearity = non_linear_map.get(non_linearity.lower(), nn.ReLU)
        
        prev = input_size
        layers = []
        for layer_size in hidden_layers:
            layers.append(nn.Linear(prev, layer_size))
            layers.append(self.non_linearity())
            prev = layer_size
        layers.pop()

        self.layers = nn.Sequential(*layers)
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class ODL(nn.Module):
    '''Flexible ODL classifier/regressor supporting classification (binary, multi-class), and regression tasks.''' 
    def __init__(self, input_size: int, output_size: int, hidden_layers: list[int],
                 non_linearity: str, b: float, s: float):
        super(ODL, self).__init__()
        
        self.register_buffer('b', torch.tensor(b, dtype=torch.float))
        self.register_buffer('s', torch.tensor(s, dtype=torch.float))
        
        # Choose non-linearity
        non_linear_map = {'relu': nn.ReLU, 'sigmoid': nn.Sigmoid, 'tanh': nn.Tanh}
        self.non_linearity = non_linear_map.get(non_linearity.lower(), nn.ReLU)()
        
        # Define hidden layers
        layer_sizes = [input_size] + hidden_layers
        self.hidden_layers = nn.ModuleList([
            nn.Linear(layer_sizes[i], layer_sizes[i+1]) for i in range(len(layer_sizes) - 1)
        ])
        
        # Define output layers corresponding to each hidden layer
        self.output_layers = nn.ModuleList([
            nn.Linear(layer_size, output_size) for layer_size in hidden_layers
        ])
        
        # Initialize alpha weights
        num_layers = max(len(hidden_layers), 1)
        self.register_buffer('alpha', torch.full((num_layers,), 1 / num_layers, dtype=torch.float))
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        hidden_outputs = []
        
        for hidden_layer, output_layer in zip(self.hidden_layers, self.output_layers):
            x = self.non_linearity(hidden_layer(x))
            hidden_outputs.append(output_layer(x).view(-1))
        
        hidden_outputs = torch.stack(hidden_outputs)  # Shape: (num_layers, output_size)
        final_pred = torch.sum(hidden_outputs * self.alpha.view(-1, 1), dim=0)  # Weighted sum
        
        return final_pred, hidden_outputs # Raw logits, activation should be applied externally
    
    def update_alpha(self, losses_per_layer: torch.Tensor):
        with torch.no_grad():
            self.alpha.mul_(torch.pow(self.b, losses_per_layer.squeeze()))
            self.alpha.copy_(torch.maximum(self.alpha, self.s / len(self.hidden_layers)))
            self.alpha.div_(self.alpha.sum())


class TaskHead(nn.Module):
    '''Getter class for various task heads available for V2F'''
    def __init__(self, head_name: str, input_size: int, output_size: int, **kwargs):
        super(TaskHead, self).__init__()
        self.name = head_name
        if head_name == 'lstm':
            self.model = LSTMModel(
                input_size=input_size, 
                output_size=output_size, 
                hidden_size=kwargs.get("hidden_size", 32), 
                num_layers=kwargs.get("num_layers", 1)
            )
        elif head_name == 'transformer':
            self.model = TransformerModel(
                input_size=input_size,
                output_size=output_size,
                num_layers=kwargs.get('num_layers', 1),
                num_heads=kwargs.get('num_heads', 4),
                dim_feedforward=kwargs.get('dim_feedforward', 32)
            )
        elif head_name == 'mlp':
            self.model = MLP(
                input_size=input_size,
                output_size=output_size,
                hidden_layers=kwargs.get('hidden_layers', [64, 64]),
                non_linearity=kwargs.get('non_linearity', 'ReLU')
            )
        elif head_name == 'odl':
            self.model = ODL(
                input_size=input_size,
                output_size=output_size,
                hidden_layers=kwargs.get('hidden_layers', [64, 64]),
                non_linearity=kwargs.get('non_linearity', 'ReLU'),
                b=kwargs.get('b', 0.99),
                s=kwargs.get('s', 0.2)
            )
        else:
            raise ValueError(f'Invalid value of `head_name` encountered: {head_name} is not supported.')
    
    def forward(self, x):
        return self.model(x)
